import math
from box import Box
import numpy as np
from scipy.spatial.transform import Rotation
import opengate_core as g4
from .generic import GenericSource, SourceBase
from ..contrib.tps.ionbeamtherapy import (
    get_spots_from_beamset_beam,
    spots_info_from_txt,
    BeamsetInfo,
)
from ..base import process_cls


def _check_ph_space_params(param_v):
    sigma = param_v[0]
    theta = param_v[1]
    epsilon = param_v[2]
    conv = param_v[3]
    pi = math.pi
    if epsilon == 0:
        raise ValueError(
            "Ellipse area is 0 !!! Check epsilon parameter in IonPencilBeamSource."
        )
    if pi * sigma * theta < epsilon:
        raise ValueError(
            f"pi*sigma*theta < epsilon. Provided values: sigma = {sigma}, theta = {theta}, epsilon = {epsilon}."
        )
    if conv not in [0, 1]:
        raise ValueError("convergence parameter can be only 0 or 1.")


class IonPencilBeamSource(GenericSource, g4.GatePencilBeamSource):
    """
    Pencil Beam source
    """

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()
        self.position.type = "disc"

        # FIXME: how to change the user_info for this new param ? -> use "override:True" entry in options dict

        # additional parameters: direction
        # sigma, theta, epsilon, conv (0: divergent, 1: convergent)
        self.direction.partPhSp_x = [0, 0, 0, 0]
        self.direction.partPhSp_y = [0, 0, 0, 0]

    def __initcpp__(self):
        g4.GatePencilBeamSource.__init__(self)

    def initialize(self, run_timing_intervals):
        GenericSource.initialize(self, run_timing_intervals)
        _check_ph_space_params(self.direction.partPhSp_x)
        _check_ph_space_params(self.direction.partPhSp_y)


class TreatmentPlanPBSource(SourceBase, g4.GateTreatmentPlanPBSource):
    """
    Treatment Plan source Pencil Beam
    """

    user_info_defaults = {
        "sorted_spot_generation": (
            False,
            {
                "doc": "If True the source will generate primaies for each spot in the plan one by one, following "
                "the order of the plan. Otherwise the spot to fire is randomly sampled at each event from a "
                "probability distribution function."
            },
        ),
        "beam_model": (
            None,
            {
                "doc": "A BeamModel object instance, containing the geometrical and energy-dependent "
                "parameters of the beam model"
            },
        ),
        "plan_path": (
            None,
            {
                "doc": "path of the treatment plan file to simulate. It can be in DICOM or Gate 9 .txt format "
            },
        ),
        "beam_data_dict": (
            None,
            {
                "doc": "If a plan path is not provided, the source can be initialized "
                "by providing custom or plan derived spot data. Check opengate.contrib.tps.ionbeamtherapy.spots_info_from_txt() "
                "for more details on the structure of this dictionary."
            },
        ),
        "beam_nr": (1, {"doc": "Which beam to simulate. Numbering starts from 1."}),
        "gantry_rot_axis": (
            "z",
            {
                "doc": "By default the source is oriented in +y direction. The source will be rotated "
                "of the gantry angle specified in the plan, around the specified axis."
            },
        ),
        "flat_generation": (
            False,
            {
                "doc": "If True, the same number of primaries is generated for each spot "
                "and the spot weight is applied to the energy deposition instead."
            },
        ),
        "particle": (
            None,
            {
                "doc": "Name of the particle generated by the source (gamma, e+ ... or an ion such as 'ion 9 18')"
            },
        ),
        "ion": (
            Box({"Z": 0, "A": 0, "E": 0}),
            {
                "doc": "If the particle is an ion, you must set Z: Atomic Number, A: Atomic Mass (nn + np +nlambda), E: Excitation energy (i.e. for metastable)"
            },
        ),
        "position": (
            Box(
                {"translation": [0, 0, 0], "rotation": Rotation.identity().as_matrix()}
            ),
            {
                "doc": "translation and rotation to be applied to the source. Note that the transform will be applied to "
                "the source AFTER the gantry rotation and the source to axis distance have been applied. Note: rotation is not used yet."
            },
        ),
        "positions": (
            [],
            {
                "doc": "list of the positions to which the single particle source (SPS) will be moved "
                "during the simulation to irradiate all the spots. Calculated internally."
            },
        ),
        "rotations": (
            [],
            {"doc": "list of the rotations of each SPS.Calculated internally."},
        ),
        "energies": ([], {"doc": "List of the SPS's energies. Calculated internally."}),
        "energy_sigmas": (
            [],
            {"doc": "List of the SPS's energy sigmas. Calculated internally."},
        ),
        "weights": (
            [],
            {"doc": "List of the weight for each spot. Calculated internally."},
        ),
        "pdf": (
            [],
            {
                "doc": "Probability distribution function. Calculated internally and used to sample the spots to irradiate."
            },
        ),
        "partPhSp_xV": (
            [],
            {
                "doc": "List of phase space parameters for each SPS. Calculated internally."
            },
        ),
        "partPhSp_yV": (
            [],
            {
                "doc": "List of phase space parameters for each SPS. Calculated internally."
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()
        # initialize internal members
        self.spots = None
        self.rotation = None
        self.translation = None
        self.d_nozzle_to_iso = None
        self.d_stear_mag_to_iso_x = None
        self.d_stear_mag_to_iso_y = None
        self.proportion_factor_x = None
        self.proportion_factor_y = None

    def __initcpp__(self):
        g4.GateTreatmentPlanPBSource.__init__(self)

    def initialize(self, run_timing_intervals):
        if not self.beam_data_dict and not self.plan_path:
            raise ValueError(
                "User must provide either a treatment plan file path or a beam data dictionary "
                "with spots and gantry angle."
            )

        # if len(self.user_info.n_primaries_vector) != len(self.user_info.run_timing_intervals):
        #     raise ValueError("Particles per run must have the same length of the number of runs")

        # set pbs param
        self._set_pbs_param_all_spots()

        # set ion param
        if self.particle.startswith("ion"):
            words = self.particle.split(" ")
            if len(words) > 1:
                self.ion.Z = words[1]
            if len(words) > 2:
                self.ion.A = words[2]
            if len(words) > 3:
                self.ion.E = words[3]

        # initialize
        SourceBase.initialize(self, run_timing_intervals)

    def get_generated_primaries(self):
        return self.g4_source.GetGeneratedPrimaries()

    def _set_pbs_param_all_spots(self):
        # initialize vectors every time, otherwise issues with MT
        positions = []
        rotations = []
        energies = []
        energy_sigmas = []
        weights = []
        partPhSp_xV = []
        partPhSp_yV = []
        beam_nr = self.beam_nr
        plan_path = self.plan_path
        gantry_rot_axis = self.gantry_rot_axis
        gantry_angle = None

        # get data from plan if provided
        if plan_path:
            if str(plan_path).endswith(".txt"):
                beam_data = spots_info_from_txt(plan_path, self.particle, beam_nr)
                self.spots = beam_data["spots"]
                gantry_angle = beam_data["gantry_angle"]
            elif str(plan_path).endswith(".dcm"):
                beamset = BeamsetInfo(plan_path)
                gantry_angle = beamset.beam_angles[beam_nr - 1]
                self.spots = get_spots_from_beamset_beam(beamset, beam_nr)
            else:
                raise ValueError(
                    "Plan path time has to be in .txt or .gantry_rot_axisdcm format"
                )
        elif self.beam_data_dict:
            self.spots = self.beam_data_dict["spots"]
            gantry_angle = self.beam_data_dict["gantry_angle"]

        # set variables for spots, to initialize pbs sources on the cpp side
        self.rotation = Rotation.from_euler(gantry_rot_axis, gantry_angle, degrees=True)
        self.translation = self.position.translation
        beamline = self.beam_model
        self.d_nozzle_to_iso = beamline.distance_nozzle_iso
        self.d_stear_mag_to_iso_x = beamline.distance_stearmag_to_isocenter_x
        self.d_stear_mag_to_iso_y = beamline.distance_stearmag_to_isocenter_y

        # mapping factors between iso center plane and nozzle plane (due to steering magnets)
        cal_proportion_factor = lambda d_magnet_iso: (
            1
            if (d_magnet_iso == float("inf"))
            else (d_magnet_iso - self.d_nozzle_to_iso) / d_magnet_iso
        )
        self.proportion_factor_x = cal_proportion_factor(self.d_stear_mag_to_iso_x)
        self.proportion_factor_y = cal_proportion_factor(self.d_stear_mag_to_iso_y)

        # probability density function
        self.pdf = self._define_pdf(flat_generation=self.flat_generation)

        for i, spot in enumerate(self.spots):
            # set energy
            energies.append(beamline.get_energy(nominal_energy=spot.energy))
            energy_sigmas.append(beamline.get_sigma_energy(nominal_energy=spot.energy))

            # position:
            positions.append(self._get_pbs_position(spot))

            # rotation:
            rotations.append(self._get_pbs_rotation(spot))

            # add weight
            if self.flat_generation:
                weights.append(spot.beamFraction * len(self.spots))
            else:
                weights.append(1.0)

            # set optics parameters
            partPhSp_xV.append(
                [
                    beamline.get_sigma_x(spot.energy),
                    beamline.get_theta_x(spot.energy),
                    beamline.get_epsilon_x(spot.energy),
                    beamline.conv_x,
                ]
            )
            partPhSp_yV.append(
                [
                    beamline.get_sigma_y(spot.energy),
                    beamline.get_theta_y(spot.energy),
                    beamline.get_epsilon_y(spot.energy),
                    beamline.conv_y,
                ]
            )
        # set vectors to user info
        self.positions = positions
        self.rotations = rotations
        self.energies = energies
        self.energy_sigmas = energy_sigmas
        self.weights = weights
        self.partPhSp_xV = partPhSp_xV
        self.partPhSp_yV = partPhSp_yV

    def _define_pdf(self, flat_generation=False):
        if flat_generation:
            pdf = [1.0 / len(self.spots) for spot in self.spots]
        else:
            pdf = [spot.beamFraction for spot in self.spots]

        # normalize vector, to assure the probabilities sum up to 1
        pdf = pdf / np.sum(pdf)

        return list(pdf)

    def _get_pbs_position(self, spot):
        # (x,y) refer to isocenter plane.
        # Need to be corrected to refer to nozzle plane
        pos = [
            spot.xiec * self.proportion_factor_x,
            spot.yiec * self.proportion_factor_y,
            self.d_nozzle_to_iso,
        ]
        # Gantry angle = 0 -> source comes from +y and is positioned along negative side of y-axis
        # https://opengate.readthedocs.io/en/latest/source_and_particle_management.html

        position = list(
            (self.rotation * Rotation.from_euler("x", np.pi / 2)).apply(pos)
            + self.translation
        )

        return position

    def _get_pbs_rotation(self, spot):
        # by default the source points in direction z+.
        # Need to account for SM direction deviation and rotation toward isocenter (270 deg around x)
        # then rotate of gantry angle
        rotation = [0.0, 0.0, 0.0]
        beta = np.arctan(spot.yiec / self.d_stear_mag_to_iso_y)
        alpha = np.arctan(spot.xiec / self.d_stear_mag_to_iso_x)
        rotation[0] = -np.pi / 2 + beta
        rotation[2] = -alpha

        # apply gantry angle
        spot_rotation = (
            self.rotation * Rotation.from_euler("xyz", rotation)
        ).as_matrix()

        return spot_rotation


process_cls(IonPencilBeamSource)
process_cls(TreatmentPlanPBSource)



import itk

import opengate_core as g4
from .generic import GenericSource
from ..image import (
    read_image_info,
    update_image_py_to_cpp,
    compute_image_3D_CDF,
)
from ..utility import ensure_filename_is_str
from ..base import process_cls


class VoxelSource(GenericSource, g4.GateVoxelSource):
    """
    VoxelSource = 3D activity distribution.
    Sampled with cumulative distribution functions.
    """

    # hints for IDE
    image: str

    user_info_defaults = {
        "image": (
            None,
            {
                "doc": "Filename of the image of the 3D activity distribution "
                "(will be automatically normalized to sum=1)",
                "is_input_file": True,
            },
        )
    }

    def __init__(self, *args, **kwargs):
        self.__initcpp__()
        super().__init__(self, *args, **kwargs)
        # the loaded image
        self.itk_image = None

    def __initcpp__(self):
        g4.GateVoxelSource.__init__(self)

    def set_transform_from_user_info(self):
        # get source image information
        src_info = read_image_info(str(self.image))
        # get the pointer to SPSVoxelPosDistribution
        pg = self.GetSPSVoxelPosDistribution()
        # update cpp image info (no need to allocate)
        update_image_py_to_cpp(self.itk_image, pg.cpp_edep_image, False)
        # set spacing
        pg.cpp_edep_image.set_spacing(src_info.spacing)
        # set origin (half size + translation and half-pixel shift)
        c = (
            -src_info.size / 2.0 * src_info.spacing
            + self.position.translation
            + src_info.spacing / 2.0
        )
        pg.cpp_edep_image.set_origin(c)

    def cumulative_distribution_functions(self):
        """
        Compute the Cumulative Distribution Function of the image
        Composed of: CDF_Z = 1D, CDF_Y = 2D, CDF_X = 3D
        """
        cdf_x, cdf_y, cdf_z = compute_image_3D_CDF(self.itk_image)

        # set CDF to the position generator
        pg = self.GetSPSVoxelPosDistribution()
        pg.SetCumulativeDistributionFunction(cdf_z, cdf_y, cdf_x)

    def initialize(self, run_timing_intervals):
        # read source image
        self.itk_image = itk.imread(ensure_filename_is_str(self.image))

        # compute position
        self.set_transform_from_user_info()

        # create Cumulative Distribution Function
        self.cumulative_distribution_functions()

        # FIXME -> check other option in position not used here

        # initialise standard options (particle energy, etc.)
        # we temporarily set the position attribute to reuse
        # the GenericSource verification
        GenericSource.initialize(self, run_timing_intervals)


process_cls(VoxelSource)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test041_dose_actor_dose_to_water", "test041"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123456
    sim.number_of_threads = 5
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    # add a material database
    # sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    # world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 10 * cm, 10 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    test_material_name = "G4_Si"
    phantom_off = sim.add_volume("Box", "phantom_off")
    phantom_off.mother = phantom.name
    phantom_off.size = [100 * mm, 20 * mm, 20 * mm]
    phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
    phantom_off.material = test_material_name
    phantom_off.color = [0, 0, 1, 1]

    # water slab
    water_slab_insert = sim.add_volume("Box", "water_slab_insert")
    water_slab_insert.mother = phantom_off.name
    water_slab_insert.size = [2 * mm, 20 * mm, 20 * mm]
    water_slab_insert.translation = [43 * mm, 0, 0]
    water_slab_insert.material = "G4_WATER"
    water_slab_insert.color = [0, 0, 1, 1]

    # si entrance
    entranceRegion = sim.add_volume("Box", "entranceRegion")
    entranceRegion.mother = phantom_off.name
    entranceRegion.size = [5 * mm, 20 * mm, 20 * mm]
    entranceRegion.translation = [47.5 * mm, 0, 0]
    entranceRegion.material = "G4_Si"
    entranceRegion.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMY"
    sim.physics_manager.global_production_cuts.all = 1000 * km
    # sim.set_cut("world", "all", 1000 * km)

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 40 * MeV
    source.particle = "proton"
    source.position.type = "disc"  # pos = Beam, shape = circle + sigma
    # rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    # source.position.radius = 8 * mm
    source.position.sigma_x = 2 * mm
    source.position.sigma_y = 2 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.n = 100

    # Define actors and collect them in a list for convenience (see below)
    dose_actors = []

    # each actor is attached to a different volume and/or scores in water/the local material
    dose_actor_IDD_d = sim.add_actor("DoseActor", "IDD_d")
    dose_actor_IDD_d.attached_to = phantom_off
    dose_actors.append(dose_actor_IDD_d)

    dose_actor_IDD_d2w = sim.add_actor("DoseActor", "IDD_d2w")
    dose_actor_IDD_d2w.attached_to = phantom_off
    dose_actor_IDD_d2w.score_in = "G4_WATER"
    dose_actors.append(dose_actor_IDD_d2w)

    dose_actor_water_slab_insert_d = sim.add_actor("DoseActor", "IDD_waterSlab_d")
    dose_actor_water_slab_insert_d.attached_to = water_slab_insert
    dose_actors.append(dose_actor_water_slab_insert_d)

    dose_actor_water_slab_insert_d2w = sim.add_actor("DoseActor", "IDD_waterSlab_d2w")
    dose_actor_water_slab_insert_d2w.attached_to = water_slab_insert
    dose_actor_water_slab_insert_d2w.score_in = "G4_WATER"
    dose_actors.append(dose_actor_water_slab_insert_d2w)

    dose_actor_entranceRegiont_d = sim.add_actor("DoseActor", "IDD_entranceRegion_d")
    dose_actor_entranceRegiont_d.attached_to = entranceRegion
    dose_actors.append(dose_actor_entranceRegiont_d)

    dose_actor_entranceRegiont_d2w = sim.add_actor(
        "DoseActor", "IDD_entranceRegion_d2w"
    )
    dose_actor_entranceRegiont_d2w.attached_to = entranceRegion
    dose_actor_entranceRegiont_d2w.score_in = "G4_WATER"
    dose_actors.append(dose_actor_entranceRegiont_d2w)

    # set common properties
    # doing this is a loop keeps this script nicely compact
    for d in dose_actors:
        d.size = [1000, 1, 1]
        d.spacing = [0.1, 20.0, 20.0]
        d.output_filename = f"test041-{d.name}.mhd"
        d.dose.active = True  # the actor only scores edep by default, so we need to active dose scoring

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    print("*** TESTING ***")

    # we can access the dose component of each actor
    # and ask it for the output path
    # so we do not need to manually keep track of the paths here in the script
    # syntax: dose_actor.dose.get_output_path()

    utility.assert_images(
        dose_actor_IDD_d.dose.get_output_path(),
        dose_actor_IDD_d2w.dose.get_output_path(),
        stats,
        tolerance=100,
        ignore_value_data2=0,
        axis="x",
    )

    mSPR_40MeV = 1.268771331  # from PSTAR NIST tables, Feb 2023
    mSPR_80MeV = 1.253197674  # from PSTAR NIST tables, Feb 2023
    gate.exception.warning(
        "Test ratio: dose / dose_to_water in geometry with material: G4_WATER"
    )
    is_ok = utility.assert_images_ratio(
        1.00,
        dose_actor_water_slab_insert_d.dose.get_output_path(),
        dose_actor_water_slab_insert_d2w.dose.get_output_path(),
        abs_tolerance=0.05,
    )

    gate.exception.warning(
        "Test ratio: dose / dose_to_water in geometry with material: G4_Si"
    )
    is_ok = (
        utility.assert_images_ratio(
            mSPR_40MeV,
            dose_actor_entranceRegiont_d.dose.get_output_path(),
            dose_actor_entranceRegiont_d2w.dose.get_output_path(),
            abs_tolerance=0.05,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test060_phsp_source_helpers as t
import opengate as gate
from opengate.tests import utility


paths = utility.get_default_test_paths(
    __file__, "test060_PhsSource_ParticleName_direct", output_folder="test060"
)

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg = gate.g4_units.deg


def main():
    print("create reference PhS file")
    t.create_test_phs(
        particle="proton",
        phs_name=paths.output / "test_proton_offset",
        number_of_particles=1,
        translation=[10 * cm, 5 * cm, 0 * mm],
    )
    print("testing translation")
    t.test_source_translation(
        source_file_name=paths.output / "test_proton_offset.root",
        phs_file_name_out=paths.output / "test_source_translation.root",
    )
    is_ok = t.check_value_from_root_file(
        file_name_root=paths.output / "test_source_translation.root",
        key="PrePositionLocal_X",
        ref_value=30 * mm,
    )

    # this is the end, my friend
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import pathlib


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test010")
    ref_path = paths.output_ref

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    deg = gate.g4_units.deg
    um = gate.g4_units.um
    keV = gate.g4_units.keV
    g_cm3 = gate.g4_units.g_cm3

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.visu_type = "vrml"
    sim.number_of_threads = 1
    sim.random_seed = 123654
    sim.output_dir = paths.output

    # set the world size like in the Gate macro
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # set daughter volume for doseactor
    image_volume = sim.add_volume("Box", "image_volume")
    image_volume.material = "G4_Pb"
    image_volume.mother = "world"
    image_volume.size = [100 * cm, 1 * cm, 100 * cm]
    image_volume.translation = [0, -50 * cm, 0]

    # test sources
    source = sim.add_source("GenericSource", "source")
    source.particle = "gamma"
    # source.n = 100 / sim.number_of_threads
    source.n = 1e6 / sim.number_of_threads
    source.position.type = "point"
    source.position.translation = [0 * cm, 0 * cm, 0 * cm]
    source.direction.type = "histogram"
    # Put zero as first value of weight
    source.direction.histogram_theta_angles = [80 * deg, 100 * deg]
    source.direction.histogram_theta_weights = [1]
    source.direction.histogram_phi_angles = [
        60 * deg,
        70 * deg,
        80 * deg,
        100 * deg,
        110 * deg,
        120 * deg,
    ]
    source.direction.histogram_phi_weights = [0.3, 0.5, 1, 0.5, 0.3]
    source.energy.type = "gauss"
    source.energy.mono = 70 * keV

    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test010-generic_source_angular_distribution.mhd"
    dose.edep_uncertainty.active = True
    dose.attached_to = "image_volume"
    dose.size = [100, 1, 100]
    dose.spacing = [10 * mm, 1 * cm, 10 * mm]

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True

    # start simulation
    sim.run()

    # get results
    print(stat)
    print(dose)

    # tests
    stats_ref = utility.read_stat_file(
        ref_path / "test010_generic_source_angular_distribution_stats_ref.txt"
    )
    is_ok = utility.assert_stats(stat, stats_ref, 0.11)

    print("\nDifference for EDEP")
    is_ok = (
        utility.assert_images(
            ref_path / "test010-generic_source_angular_distribution_edep_ref.mhd",
            dose.edep.get_output_path(),
            stat,
            tolerance=13,
            ignore_value_data2=0,
            sum_tolerance=1,
        )
        and is_ok
    )

    print("\nDifference for uncertainty")
    is_ok = (
        utility.assert_images(
            ref_path
            / "test010-generic_source_angular_distribution_edep_uncertainty_ref.mhd",
            dose.edep_uncertainty.get_output_path(),
            stat,
            tolerance=30,
            ignore_value_data2=0,
            sum_tolerance=1,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from box import Box
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test045")

    p = Box()
    p.phantom_type = "analytic"
    p.source_type = "analytic"
    p.use_pet = True
    p.use_gaga = False
    p.a = 1e2

    # debug
    skip = False
    # the seed is only for the G4 part, the GAN use his own seed
    # so the results are not reproducible (yet)
    seed = 812365478

    def run(param):
        print("run ", param)
        cmd_line = (
            f"{paths.current}/test045_speedup.py -o AUTO --seed {seed} -p {param.phantom_type} "
            f"-s {param.source_type} -r Ga68 -a {param.a}"
        )
        if param.use_pet:
            cmd_line += " --pet "
        if param.use_gaga:
            cmd_line += " --gaga "
        out = f"test045_speedup_p_{param.phantom_type}_s_{param.source_type}_pet_{param.use_pet}_gaga_{param.use_gaga}.txt"
        if not skip:
            print("cmd line", cmd_line)
            r = os.system(f"python {cmd_line}")

        print("Output ", out)
        return out

    # output
    output = []

    # Test 1
    p.phantom_type = "analytic"
    p.source_type = "analytic"
    p.use_gaga = False
    p.a = 1e3
    out = run(p)
    output.append(out)

    # Test 2
    p.phantom_type = "analytic"
    p.source_type = "vox"
    p.use_gaga = False
    p.a = 1e3
    out = run(p)
    output.append(out)

    # Test 3
    p.phantom_type = "vox"
    p.source_type = "vox"
    p.use_gaga = False
    p.a = 1e3
    out = run(p)
    output.append(out)

    # Test 4
    p.phantom_type = "analytic"
    p.source_type = "analytic"
    p.use_gaga = True
    p.a = 1e3
    out = run(p)
    output.append(out)

    # Test 5
    p.phantom_type = "analytic"
    p.source_type = "vox"
    p.use_gaga = True
    p.a = 1e3
    out = run(p)
    output.append(out)

    # Test 6
    p.phantom_type = "vox"
    p.source_type = "vox"
    p.use_gaga = True
    p.a = 1e3
    out = run(p)
    output.append(out)

    print(output)

    # tests stats file
    is_ok = True
    for o in output:
        stats = utility.read_stat_file(paths.output / o)
        stats_ref = utility.read_stat_file(paths.output_ref / o)
        ok = utility.assert_stats(stats, stats_ref, 0.06)
        utility.print_test(ok, f"Check {o}")
        is_ok = is_ok and ok
        print()

    # tests pet files
    keys = [
        "GlobalTime",
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
    ]
    tols = [10.0] * len(keys)
    tols[keys.index("GlobalTime")] = 0.04
    tols[keys.index("PostPosition_X")] = 8.4
    tols[keys.index("PostPosition_Y")] = 13
    tols[keys.index("PostPosition_Z")] = 1.5
    tols[keys.index("TotalEnergyDeposit")] = 0.03
    tols[keys.index("TrackVolumeCopyNo")] = 4.1
    scalings = [1.0] * len(keys)
    scalings[keys.index("GlobalTime")] = 1e-9  # time in ns
    for o in output:
        o = o.replace(".txt", ".root")
        o1 = paths.output / o
        o2 = paths.output_ref / o
        img = paths.output / o.replace(".root", ".png")
        ok = utility.compare_root3(
            o1,
            o2,
            "Singles",
            "Singles",
            keys,
            keys,
            tols,
            scalings,
            scalings,
            img,
            hits_tol=5,
        )
        utility.print_test(ok, f"Check {o}")
        is_ok = is_ok and ok
        print()

    utility.test_ok(is_ok)



from pathlib import Path
import json
import random
import string
import itk
import numpy as np

import opengate_core as g4
from .engines import SimulationEngine
from .exception import fatal
from .geometry.volumes import VolumeBase
from .image import (
    write_itk_image,
    create_image_with_volume_extent,
    create_image_with_extent,
    update_image_py_to_cpp,
    get_py_image_from_cpp_image,
    get_info_from_image,
)
from .processing import dispatch_to_subprocess
from .serialization import dump_json
from .utility import ensure_filename_is_str
from .definitions import __gate_list_objects__
from . import logger


def generate_random_string(length=10):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def voxelize_geometry(
    sim,
    extent="auto",
    spacing=(3, 3, 3),
    margin=0,
    filename=None,
    return_path=False,
):
    """Create a voxelized three-dimensional representation of the simulation geometry.

    The user can specify the sub-portion (a rectangular box) of the simulation which is to be extracted.

    Args:
        extent : By default ('auto'), GATE automatically determines the sub-portion
            to contain all volumes of the simulation.
            Alternatively, extent can be either a tuple of 3-vectors indicating the two diagonally
            opposite corners of the box-shaped
            sub-portion of the geometry to be extracted, or a volume or list volumes.
            In the latter case, the box is automatically determined to contain the volume(s).
        spacing (tuple): The voxel spacing in x-, y-, z-direction.
        margin : Width (in voxels) of the additional margin around the extracted box-shaped sub-portion
            indicated by `extent`.
        filename (str, optional): The filename/path to which the voxelized image and labels are written.
            Suffix added automatically. Path can be relative to the global output directory of the simulation.
        return_path (bool): Return the absolute path where the voxelized image was written?

    Returns:
        dict, itk image, (path): A dictionary containing the label to volume LUT; the voxelized geometry;
            optionally: the absolute path where the image was written, if applicable.
    """
    # collect volumes which are directly underneath the world/parallel worlds
    if extent in ("auto", "Auto"):
        sim.volume_manager.update_volume_tree_if_needed()
        extent = list(sim.volume_manager.world_volume.children)
        for pw in sim.volume_manager.parallel_world_volumes.values():
            extent.extend(list(pw.children))

    labels, image = dispatch_to_subprocess(
        compute_voxelized_geometry, sim, extent, spacing, margin
    )

    if filename is not None:
        outpath = sim.get_output_path(filename)
        outpath_json = outpath.parent / (outpath.stem + "_labels.json")
        outpath_mhd = outpath.parent / (outpath.stem + "_image.mhd")

        # write labels
        with open(outpath_json, "w") as outfile:
            dump_json(labels, outfile, indent=4)

        # write image
        write_itk_image(image, ensure_filename_is_str(outpath_mhd))
    else:
        outpath_mhd = "not_applicable"

    if return_path is True:
        return labels, image, outpath_mhd
    else:
        return labels, image


def write_voxelized_geometry(
    self,
    labels,
    image,
    base_filename,
    vol_filename=None,
    image_filename=None,
    label_filename=None,
    db_filename=None,
):
    # write labels
    if vol_filename is None:
        vol_filename = Path(base_filename).with_suffix(".json")
        vol_filename = str(vol_filename).replace(".json", "_volumes.json")
    with open(vol_filename, "w") as outfile:
        json.dump(labels, outfile, indent=4)

    # write image
    if image_filename is None:
        image_filename = Path(base_filename).with_suffix(".mhd")
    itk.imwrite(image, image_filename)

    # create the label to material
    vm = [[m["label"], m["label"] + 1, m["material"]] for m in labels.values()]
    image_volume = self.add_volume("Image", generate_random_string(12))
    image_volume.voxel_materials = vm
    if label_filename is None:
        label_filename = Path(base_filename).with_suffix(".json")
        label_filename = str(label_filename).replace(".json", "_labels.json")
    image_volume.write_label_to_material(label_filename)

    # write the database of material
    # gate_iec.create_material(sim)
    if db_filename is None:
        db_filename = Path(base_filename).with_suffix(".db")
    image_volume.write_material_database(db_filename)

    self.volume_manager.remove_volume(image_volume.name)
    return {
        "volumes": vol_filename,
        "image": image_filename,
        "labels": label_filename,
        "materials": db_filename,
    }


def compute_voxelized_geometry(sim, extent, spacing, margin):
    """Method which returns a voxelized image of the simulation geometry
    given the extent, spacing and margin.
    The voxelization does not check which volume is voxelized.
    Every voxel will be assigned an ID corresponding to the material at this position
    in the world.
    """

    if isinstance(extent, VolumeBase):
        image = create_image_with_volume_extent(extent, spacing, margin)
    elif isinstance(extent, __gate_list_objects__) and all(
        [isinstance(e, VolumeBase) for e in extent]
    ):
        image = create_image_with_volume_extent(extent, spacing, margin)
    elif isinstance(extent, __gate_list_objects__) and all(
        [isinstance(e, __gate_list_objects__) and len(e) == 3 for e in extent]
    ):
        image = create_image_with_extent(extent, spacing, margin)
    else:
        fatal(
            f"The input variable `extent` needs to be a tuple of 3-vectors, or a volume, "
            f"or a list of volumes. Found: {extent}."
        )

    vl = sim.verbose_level
    sim.verbose_level = logger.NONE
    with SimulationEngine(sim) as se:
        se.initialize()
        vox = g4.GateVolumeVoxelizer()
        update_image_py_to_cpp(image, vox.fImage, False)
        vox.Voxelize()
        image = get_py_image_from_cpp_image(vox.fImage)
        labels = vox.fLabels
        for key in labels.keys():
            vol = se.simulation.volume_manager.get_volume(key)
            labels[key] = {"label": labels[key], "material": vol.material}

    sim.verbose_level = vl
    return labels, image


def voxelized_source(itk_image, volumes_labels, activities):
    img_label = itk.GetArrayViewFromImage(itk_image)
    img_arr = itk.GetArrayFromImage(itk_image).astype(np.float32)
    img_arr[:, :, :] = 0.0
    for label in volumes_labels:
        l = volumes_labels[label]["label"]
        if label in activities:
            img_arr[img_label == l] = activities[label]
    itk_source = itk.GetImageFromArray(img_arr)
    itk_source.CopyInformation(itk_image)
    return itk_source



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test025_hits_collection_helpers as t025

if __name__ == "__main__":
    sim = t025.create_simulation(1)
    sim.run()
    t025.test_simulation_results(sim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation
from pathlib import Path

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test008_dose_actor")
    ref_path = paths.gate_output

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12345678
    sim.output_dir = paths.output / Path(__file__.rstrip(".py")).stem

    # shortcuts for units
    m = gate.g4_units.m
    cm = gate.g4_units.cm

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.rotation = Rotation.from_euler("x", 10, degrees=True).as_matrix()
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", 20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.apply_cuts = True  # default
    um = gate.g4_units.um
    global_cut = 700 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    source.energy.mono = 150 * MeV
    nm = gate.g4_units.nm
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 1 * nm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 50000 * Bq

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.attached_to = "waterbox"
    dose.size = [99, 99, 99]
    mm = gate.g4_units.mm
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.edep_uncertainty.active = True
    dose.edep_squared.active = True
    dose.hit_type = "random"
    dose.output_coordinate_system = "local"
    dose.output_filename = "test.nii.gz"
    dose.edep.keep_data_per_run = True

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True

    sim.run_timing_intervals = [
        (x / 3.0 * gate.g4_units.s, (x + 1) / 3.0 * gate.g4_units.s) for x in range(3)
    ]
    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stat)
    print(dose)

    # tests
    stats_ref = utility.read_stat_file(ref_path / "stat.txt")
    stat.counts.runs = 1  # because ref had only 1 run
    is_ok = utility.assert_stats(stat, stats_ref, 0.11)

    print("\nDifference for EDEP")
    is_ok = (
        utility.assert_images(
            ref_path / "output-Edep.mhd",
            dose.edep.get_output_path(),
            stat,
            tolerance=13,
            ignore_value_data2=0,
            sum_tolerance=2.5,
        )
        and is_ok
    )

    print("\nDifference for Square")
    is_ok = (
        utility.assert_images(
            ref_path / "output-Edep-Squared.mhd",
            dose.edep_squared.get_output_path(),
            stat,
            tolerance=8,
            ignore_value_data2=0,
            sum_tolerance=1.5,
        )
        and is_ok
    )

    print("\nDifference for uncertainty")
    is_ok = (
        utility.assert_images(
            ref_path / "output-Edep-Uncertainty.mhd",
            dose.edep_uncertainty.get_output_path(),
            stat,
            tolerance=30,
            ignore_value_data2=0,
            sum_tolerance=3,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



import os
import sys
import site


def get_site_packages_dir():
    site_package = [p for p in site.getsitepackages() if "site-packages" in p]
    if len(site_package) > 0:
        site_package = site_package[0]
    else:
        site_package = site.getusersitepackages()
    return site_package


def get_lib_g4_path(lib):
    for element in os.listdir(
        os.path.join(get_site_packages_dir(), "opengate_core.libs")
    ):
        if "libG4" + lib in element:
            return os.path.join(get_site_packages_dir(), "opengate_core.libs", element)


tunables_value = "glibc.rtld.optional_static_tls=2000000"


def print_ld_preload_error(developer_mode=False):
    print("=" * 80)
    print("The opengate_core library cannot be loaded.")
    print("Error is: cannot allocate memory in static TLS block")
    print("Please use the following export lines before importing opengate:")
    print(
        "export LD_LIBRARY_PATH="
        + os.path.join(get_site_packages_dir(), "opengate_core.libs")
        + ":${LD_LIBRARY_PATH}"
    )
    if developer_mode:
        print(
            "export LD_PRELOAD=<path_to_libG4processes.so>:<path_to_libG4processes.so>:${LD_LIBRARY_PATH}"
        )
    else:
        print(
            "export LD_PRELOAD="
            + get_lib_g4_path("processes")
            + ":"
            + get_lib_g4_path("geometry")
        )
    print(f"or: \n" f"export GLIBC_TUNABLES={tunables_value}")
    print("=" * 80)
    print()


def is_python_interactive_shell():
    import __main__

    return not hasattr(__main__, "__file__")


def is_developer_installation():
    path_current_file = os.path.abspath(__file__)
    if "site-packages" in path_current_file:
        return False
    return True


def restart_with_qt_libs():
    """
    Restart the current process with QT libs set.
    Only for mac and wheel installation
    """
    developer_mode = is_developer_installation()
    if is_python_interactive_shell() or developer_mode:
        # we cannot know the real plugin_path nor if DYLD_LIBRARY_PATH is already correctly set
        # so we ignore.
        # no vizualisation
        return

    plugin_path = os.path.join(get_site_packages_dir(), "opengate_core/plugins")

    # do nothing if the plugin_path is already in the env
    if (
        "DYLD_LIBRARY_PATH" in os.environ
        and plugin_path in os.environ["DYLD_LIBRARY_PATH"]
    ):
        return

    # Otherwise, we set the env and try to restart the script
    new_env = os.environ.copy()
    if "DYLD_LIBRARY_PATH" in new_env:
        new_env["DYLD_LIBRARY_PATH"] = plugin_path + new_env["DYLD_LIBRARY_PATH"]
    else:
        new_env["DYLD_LIBRARY_PATH"] = plugin_path

    # Restart the process with the new environment
    # print(new_env["DYLD_LIBRARY_PATH"])
    os.execve(sys.executable, [sys.executable] + sys.argv, new_env)


def check_ld_preload(e):
    """
    Check if the LD_PRELOAD environment variable is set correctly.
    """

    # Check if opengate_core can be loaded, if yes we continue
    if "cannot allocate memory in static TLS block" not in str(e):
        print("Cannot import opengate_core module")
        print(e)
        exit(-1)

    developer_mode = is_developer_installation()

    # We can do nothing if this is an interactive shell
    if is_python_interactive_shell():
        print_ld_preload_error(developer_mode)
        exit(-1)

    # print error message
    if developer_mode:
        print_ld_preload_error(developer_mode)
        exit(-1)


try:
    import opengate_core
except ImportError as e:
    if sys.platform.startswith("linux"):
        check_ld_preload(e)

if sys.platform.startswith("darwin"):
    restart_with_qt_libs()

# subpackages
import opengate.sources
import opengate.geometry
import opengate.geometry.materials
import opengate.geometry.solids
import opengate.geometry.utility
import opengate.geometry.volumes
import opengate.actors
import opengate.contrib

# modules directly under /opengate/
import opengate.managers
import opengate.utility
import opengate.logger
import opengate.exception
import opengate.runtiming
import opengate.definitions
import opengate.userhooks
import opengate.image
import opengate.physics
import opengate.base
import opengate.engines

# import opengate.postprocessors

# These objects are imported at the top level of the package
# because users will frequently use them
from opengate.managers import Simulation
from opengate.managers import create_sim_from_json
from opengate.utility import g4_units
from opengate.base import help_on_user_info



from opengate.contrib.phantoms.nemaiec import dump_bg_activity
from opengate.utility import g4_units


def add_phantom_material(simulation):
    elems = ["H", "C", "O"]
    w = [0.080538, 0.599848, 0.319614]
    gcm3 = g4_units.g_cm3
    simulation.volume_manager.material_database.add_material_weights(
        "JASZCZAK_PMMA", elems, w, 1.19 * gcm3
    )


def add_jaszczak_phantom(simulation, name="jaszczak", check_overlap=False):
    # https://www.meditest.fr/documentation/4.pdf
    # unit
    add_phantom_material(simulation)

    # check overlap only for debug
    original_check_overlap_flag = simulation.check_volumes_overlap
    simulation.check_volumes_overlap = check_overlap

    # Outside structure
    jaszczak, internal_cylinder = add_jaszczak_body(simulation, name)
    jaszczak.material = "JASZCZAK_PMMA"

    simulation.check_volumes_overlap = original_check_overlap_flag
    return jaszczak, internal_cylinder


def add_jaszczak_body(sim, name):
    cm = g4_units.cm
    mm = g4_units.mm
    deg = g4_units.deg
    white = [1, 1, 1, 1]
    thickness = 3.2 * mm

    # main cylinder interior
    external_cylinder = sim.add_volume("Tubs", name=f"{name}_cylinder")
    external_cylinder.rmax = 21.6 * cm / 2.0 + thickness
    external_cylinder.rmin = 0
    external_cylinder.dz = 18.6 * cm / 2 + thickness * 2
    external_cylinder.sphi = 0 * deg
    external_cylinder.dphi = 360 * deg
    external_cylinder.material = "JASZCZAK_PMMA"
    external_cylinder.color = white

    # main cylinder interior
    internal_cylinder = sim.add_volume("Tubs", name=f"{name}_internal_cylinder")
    internal_cylinder.mother = external_cylinder.name
    internal_cylinder.rmax = 21.6 * cm / 2
    internal_cylinder.rmin = 0
    internal_cylinder.dz = 18.6 * cm / 2
    internal_cylinder.sphi = 0 * deg
    internal_cylinder.dphi = 360 * deg
    internal_cylinder.material = "G4_AIR"

    return external_cylinder, internal_cylinder


def add_background_source(
    simulation, jaszczak_name, src_name, activity_bqml, verbose=False
):
    # source
    bg = simulation.add_source("GenericSource", f"{jaszczak_name}_{src_name}")
    bg.attached_to = f"{jaszczak_name}_internal_cylinder"
    v = simulation.volume_manager.volumes[bg.attached_to]
    v_info = v.solid_info
    # (1 cm3 = 1 mL)
    """
    # no need to confine yet, nothing inside the cylinder -> should be done later
    bg.position.type = "box"
    bg.position.size = simulation.volume_manager.volumes[bg.mother].bounding_box_size
    # this source is confined only within the mother volume, it does not include daughter volumes
    # it is a tubs inside the box
    bg.position.confine = bg.mother
    """
    bg.particle = "e+"
    bg.energy.type = "F18"
    bg.activity = activity_bqml * v_info.cubic_volume

    if verbose:
        s = dump_bg_activity(simulation, jaszczak_name, src_name)
        print(s)
    return bg, v_info.cubic_volume



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import numpy as np
import opengate_core as g4
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test053_digit_efficiency", "test057"
    )

    """
    PET simulation to test efficiency options of the digitizer
    Output: singles with and without decreased efficiency
    """

    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 1
    sim.check_volumes_overlap = False
    sim.random_seed = 321654
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq

    # world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # material
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # fake spect head
    waterbox = sim.add_volume("Box", "SPECThead")
    waterbox.size = [55 * cm, 42 * cm, 18 * cm]
    waterbox.material = "G4_AIR"

    # crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = "SPECThead"
    crystal.size = [1.0 * cm, 1.0 * cm, 1.0 * cm]
    crystal.material = "NaITl"
    start = [-25 * cm, -20 * cm, 4 * cm]
    size = [100, 40, 1]
    # size = [100, 80, 1]
    tr = [0.5 * cm, 0.5 * cm, 0]
    crystal.translation = gate.geometry.utility.get_grid_repetition(
        size, tr, start=start
    )
    crystal.color = [1, 1, 0, 1]

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 0.01 * mm
    sim.physics_manager.global_production_cuts.electron = 0.01 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    sim.physics_manager.global_production_cuts.proton = 1 * mm

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 4 * cm
    source.position.translation = [0, 0, -15 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 50000 * Bq / sim.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # print list of attributes
    am = g4.GateDigiAttributeManager.GetInstance()
    print(am.GetAvailableDigiAttributeNames())

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = crystal.name
    mt = ""
    if sim.number_of_threads > 1:
        mt = "_MT"
    hc.output_filename = f"test053_hits{mt}.root"
    hc.attributes = [
        "TotalEnergyDeposit",
        "KineticEnergy",
        "PostPosition",
        "TrackCreatorProcess",
        "GlobalTime",
        "TrackVolumeName",
        "RunID",
        "ThreadID",
        "TrackID",
    ]

    # EfficiencyActor
    ea = sim.add_actor("DigitizerEfficiencyActor", "Efficiency")
    ea.attached_to = hc.attached_to
    ea.input_digi_collection = "Hits"
    ea.output_filename = hc.output_filename
    ea.efficiency = 0.3

    # go
    sim.run()

    # Compare Hits and Efficiency
    hits1 = uproot.open(hc.get_output_path())["Hits"]
    hits1_n = hits1.num_entries
    hits1 = hits1.arrays(library="numpy")

    hits2 = uproot.open(hc.get_output_path())["Efficiency"]
    hits2_n = hits2.num_entries
    hits2 = hits2.arrays(library="numpy")

    print(f"Reference tree: Hits       n={hits1_n}")
    print(f"Current tree:   Efficiency n={hits2_n}")
    print(f"Digitizer efficiency = {ea.efficiency}")
    n_tol = 1.1
    diff = utility.rel_diff(float(hits1_n * ea.efficiency), float(hits2_n))
    is_ok = utility.print_test(
        np.fabs(diff) < n_tol,
        f"Difference: {ea.efficiency}*{hits1_n} {hits2_n} {diff:.2f}% (tol = {n_tol:.2f})",
    )
    print(f"Reference tree: {hits1.keys()}")
    print(f"Current tree:   {hits2.keys()}")

    keys1, keys2, scalings, tols = utility.get_keys_correspondence(
        [
            "TotalEnergyDeposit",
            "KineticEnergy",
            "PostPosition_X",
            "PostPosition_Y",
            "PostPosition_Z",
            "TrackCreatorProcess",
            "GlobalTime",
            "TrackVolumeName",
            "RunID",
            "ThreadID",
            "TrackID",
        ]
    )
    is_ok = (
        utility.compare_trees(
            hits1,
            list(hits1.keys()),
            hits2,
            list(hits2.keys()),
            keys1,
            keys2,
            tols,
            [1] * len(scalings),
            scalings,
            False,
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility


def run_test_re(particle="carbon"):

    paths = utility.get_default_test_paths(__file__, "test_087")
    print(paths)
    ref_path = paths.output_ref / "test087"
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.random_seed = 12345678910
    ui.number_of_threads = 1
    if particle == "proton":
        numPartSimTest = 5e2 / ui.number_of_threads
    else:
        numPartSimTest = 4000 / ui.number_of_threads

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    km = gate.g4_units.km
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    # world.material = "Vacuum"

    # waterbox
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [10 * cm, 10 * cm, 10 * cm]
    phantom.translation = [-5 * cm, 0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    test_material_name = "G4_WATER"
    phantom_off = sim.add_volume("Box", "phantom_off")
    phantom_off.mother = phantom.name
    phantom_off.size = [100 * mm, 60 * mm, 60 * mm]
    phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
    phantom_off.material = test_material_name
    phantom_off.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_HP_EMZ"

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    if particle == "proton":
        source.energy.mono = 80 * MeV
        source.particle = "proton"
    else:
        source.energy.mono = 1424 * MeV
        source.particle = "ion 6 12"

    source.position.type = "disc"
    source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    source.position.radius = 4 * mm
    source.position.translation = [0, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    # print(dir(source.energy))
    source.n = numPartSimTest
    # source.activity = 100 * kBq

    size = [50, 1, 1]
    spacing = [2.0 * mm, 60.0 * mm, 60.0 * mm]

    doseActorName_IDD_d = f"IDD_d_{particle}"
    doseIDD = sim.add_actor("DoseActor", doseActorName_IDD_d)
    doseIDD.output_filename = paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")
    #    print(f'actor: {paths.output / ("test087-" + doseActorName_IDD_d + ".mhd")}')
    doseIDD.attached_to = phantom_off.name
    doseIDD.size = size
    doseIDD.spacing = spacing
    doseIDD.hit_type = "random"
    doseIDD.dose.active = False

    RE = f"RE_{particle}"
    RE_act = sim.add_actor("REActor", RE)
    RE_act.output_filename = paths.output / ("test087-" + RE + ".mhd")
    RE_act.attached_to = phantom_off.name
    RE_act.size = size
    RE_act.spacing = spacing
    RE_act.hit_type = "random"
    RE_act.model = "RE"
    # TODO: note that EMCalculator throws a segfault for O16 at low energies for G4_Alanine; therfore we cannot use it although we should
    #    RE_act.score_in = "G4_ALANINE"
    RE_act.score_in = "G4_WATER"
    RE_act.lookup_table_path = ref_path / "RE_Alanine_RBEstyle.txt"

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True

    sim.run(start_new_process=True)

    # ----------------------------------------------------------------------------------------------------------------

    if particle == "proton":
        ref_fpath = (
            ref_path
            / "test087_REalanine__Proton_Energy80spread1MeV_PrimaryProton-relEfficiency-letToG4_ALANINE.mhd"
        )
    else:
        ref_fpath = (
            ref_path
            / "test087_REalanine__Carbon_Energy1424spread1MeV_PrimaryCarbon-relEfficiency-letToG4_ALANINE.mhd"
        )
    print(f"{doseIDD.dose.get_output_path()=}")
    is_ok = utility.assert_filtered_imagesprofile1D(
        ref_filter_filename1=doseIDD.edep.get_output_path(),
        ref_filename1=ref_fpath,
        filename2=paths.output / RE_act.RE_mix.get_output_path(),
        tolerance=20,
        #        plt_ylim=[0, 2],
        eval_quantity="RE",
    )
    return is_ok


def main():
    print("Running proton test case for relative effectiveness")
    is_ok = run_test_re("proton")
    print("... continue with carbon ion beam")
    is_ok = is_ok and run_test_re("carbon")
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
#   Copyright (C): MedAustron GmbH, ACMIT Gmbh and Medical University Vienna
#   This software is distributed under the terms
#   of the GNU Lesser General  Public Licence (LGPL)
#   See LICENSE for further details
# -----------------------------------------------------------------------------

import pydicom
import os
import re
import numpy as np
import logging
from scipy.spatial.transform import Rotation
import opengate as gate

logger = logging.getLogger(__name__)


def is_close(x, y, eps=1e-6):
    sumabs = np.abs(x) + np.abs(y)
    absdif = np.abs(x - y)
    ok = (sumabs == 0) or (absdif < eps * 0.5 * sumabs)
    return ok


def sequence_check(obj, attr, nmin=1, nmax=0, name="object"):
    logger.debug("checking that {} has attribute {}".format(name, attr))
    assert hasattr(obj, attr)
    seq = getattr(obj, attr)
    logger.debug(
        "{} has length {}, will check if it >={} and <={}".format(
            name, len(seq), nmin, nmax
        )
    )
    assert len(seq) >= nmin
    assert nmax == 0 or len(seq) <= nmax


def spots_info_from_txt(txtFile, ionType, beam_nr):
    # initialize empty variables

    beam_data = dict()
    beam_data["n_fields"] = 0
    beam_data["plan_name"] = ""
    beam_data["msw_beam"] = 0
    beam_data["energies"] = []
    beam_data["nb_spots"] = []
    beam_data["spots"] = []
    beam_data["gantry_angle"] = 0
    beam_data["couch_angle"] = 0
    beam_data["isocenter"] = []
    found_field = False
    start_index = []

    # read  content
    with open(txtFile, "r") as f:
        lines = f.readlines()

    # get plan's info
    # TODO: make function to check line tag
    for i, line in enumerate(lines):
        if check_plan_tag(line, "PlanName"):
            l = lines[i + 1].split("\n")[0]
            beam_data["plan_name"] = l
        if check_plan_tag(line, "NumberOfFields"):
            l = lines[i + 1].split("\n")[0]
            beam_data["n_fields"] = int(l)
            if beam_nr > beam_data["n_fields"]:

                raise ValueError(
                    "requested beam number higher than number of beams in the beamset"
                )
        if check_plan_tag(line, "FIELD-DESCRIPTION"):
            found_field = False
        if check_plan_tag(line, "FieldID"):
            fieldID = int(lines[i + 1].split("\n")[0])
            if fieldID == beam_nr:
                found_field = True
        if found_field:
            if check_plan_tag(line, "GantryAngle"):
                l = lines[i + 1].split("\n")[0]
                beam_data["gantry_angle"] = float(l)
            if check_plan_tag(line, "PatientSupportAngle"):
                l = lines[i + 1].split("\n")[0]
                beam_data["couch_angle"] = float(l)
            if check_plan_tag(line, "IsocenterPosition"):
                l = lines[i + 1].split("\n")[0]
                beam_data["isocenter"] = [float(i) for i in l.split()]
            if check_plan_tag(line, "FinalCumulativeMeterSetWeight"):
                l = lines[i + 1].split("\n")[0]
                beam_data["msw_beam"] = float(l)
            if check_plan_tag(line, "Energy"):
                l = lines[i + 1].split("\n")[0]
                beam_data["energies"].append(float(l))
            if check_plan_tag(line, "NbOfScannedSpots"):
                l = lines[i + 1].split("\n")[0]
                beam_data["nb_spots"].append(int(l))
            if check_plan_tag(line, "X Y Weight"):
                start_index.append(i + 1)

    for i in range(len(beam_data["energies"])):
        e = beam_data["energies"][i]
        # print(f"ENERGY: {e}")
        start = start_index[i]
        end = start_index[i] + beam_data["nb_spots"][i]
        for j in range(start, end):
            l = lines[j].split("\n")[0].split()
            spot = SpotInfo(float(l[0]), float(l[1]), float(l[2]), e)
            spot.beamFraction = float(l[2]) / beam_data["msw_beam"]
            spot.particle_name = ionType
            beam_data["spots"].append(spot)

    return beam_data


def check_plan_tag(txt_line, tag):
    txt_line = txt_line.strip().lower()
    tag = tag.strip().lower()
    return tag in txt_line


def get_spots_from_beamset(beamset):
    rad_type = beamset.bs_info["Radiation Type Opengate"]
    spots_array = []
    mswtot = beamset.mswtot
    for beam in beamset.beams:
        # mswtot = beam.mswtot
        for energy_layer in beam.layers:
            for spot in energy_layer.spots:
                nPlannedSpot = spot.w
                spot.beamFraction = (
                    nPlannedSpot / mswtot
                )  # nr particles planned for the spot/tot particles planned for the beam
                spot.particle_name = rad_type
                spots_array.append(spot)
    return spots_array


def get_spots_from_beamset_beam(beamset, beam_nr):
    rad_type = beamset.bs_info["Radiation Type Opengate"]
    spots_array = []
    beam = beamset.beams[beam_nr - 1]
    mswtot = beam.mswtot
    for energy_layer in beam.layers:
        for spot in energy_layer.spots:
            nPlannedSpot = spot.w
            spot.beamFraction = (
                nPlannedSpot / mswtot
            )  # nr particles planned for the spot/tot particles planned for the beam
            spot.particle_name = rad_type
            spots_array.append(spot)
    return spots_array


# FIXME: IonSpotInfo
class SpotInfo(object):
    def __init__(self, xiec, yiec, w, e):
        self.xiec = xiec
        self.yiec = yiec
        self.w = w
        self.energy = e
        self.particle_name = None
        self.beamFraction = None
        self.t0 = None
        self.t1 = None

    def get_msw(self, t0, t1):
        return self.w


class LayerInfo(object):
    def __init__(self, ctrlpnt, j, cumsumchk=[], verbose=False, keep0=False):
        self._cp = ctrlpnt
        if verbose:
            logger.debug("{}. control point with type {}".format(j, type(self._cp)))
            for k in self._cp.keys():
                if pydicom.datadict.dictionary_has_tag(k):
                    kw = pydicom.datadict.keyword_for_tag(k)
                else:
                    kw = "(UNKNOWN)"
                logger.debug("k={} keyword={}".format(k, kw))
        nspot = int(self._cp.NumberOfScanSpotPositions)
        # assert(self._cp.NominalBeamEnergyUnit == 'MEV')
        if nspot == 1:
            self.w = np.array([float(self._cp.ScanSpotMetersetWeights)])
        else:
            self.w = np.array([float(w) for w in self._cp.ScanSpotMetersetWeights])
        assert nspot == len(self.w)
        assert nspot * 2 == len(self._cp.ScanSpotPositionMap)
        # self.cpindex = int(self._cp.ControlPointIndex)
        # self.spotID = str(self._cp.ScanSpotTuneID)
        cmsw = float(self._cp.CumulativeMetersetWeight)
        if cumsumchk:
            logger.debug(
                "CumulativeMetersetWeight={0:14.8g} sum of previous spots={1:14.8g} diff={2:14.8g}".format(
                    cmsw, cumsumchk[0], cmsw - cumsumchk[0]
                )
            )
            assert is_close(cmsw, cumsumchk[0])
        # self.npainting = int(self._cp.NumberOfPaintings)
        xy = np.array([float(pos) for pos in self._cp.ScanSpotPositionMap]).reshape(
            nspot, 2
        )
        self.x = np.array(xy[:, 0])
        self.y = np.array(xy[:, 1])
        if not keep0:
            mask = self.w > 0.0
            self.w = self.w[mask]
            self.x = self.x[mask]
            self.y = self.y[mask]

        wsum = np.sum(self.w)
        logger.debug(
            "layer number {} has {} spots, of which {} with zero weight, cumsum={}, sum(w)={}".format(
                j, len(self.w), np.sum(self.w <= 0), cmsw, wsum
            )
        )

        cumsumchk[0] += wsum

    @property
    def energy(self):
        # DICOM specifies energy per nucleon, Gate wants total kinetic energy
        return float(self._cp.NominalBeamEnergy)

    @property
    def tuneID(self):
        return str(self._cp.ScanSpotTuneID)

    @property
    def npainting(self):
        return int(self._cp.NumberOfPaintings)

    @property
    def mswtot(self):
        return np.sum(self.w)

    @property
    def nspots(self):
        return len(self.w)

    @property
    def weights(self):
        return self.w

    @property
    def spots(self):
        e = self.energy
        return [SpotInfo(x, y, w, e) for (x, y, w) in zip(self.x, self.y, self.w)]

    def get_spots(self, t0=None, t1=None):
        e = self.energy
        return [SpotInfo(x, y, w, e) for (x, y, w) in zip(self.x, self.y, self.w)]


class BeamInfo(object):
    # def __init__(self,beam,rd,i,keep0=False):
    def __init__(self, beam, i, override_number, keep0=False):
        logger.debug("loading {}th beam".format(i))
        self._dcmbeam = beam  # the DICOM beam object
        self._warnings = list()  # will hopefully stay empty
        logger.debug("trying to access first control point")
        self._icp0 = beam.IonControlPointSequence[0]  # convenience: first control point
        self._beam_number_is_fishy = (
            override_number  # workaround for buggy TPSs, e.g. PDM
        )
        self._index = i  # the index in the beam sequence
        self._layers = list()
        mswchk = self.FinalCumulativeMetersetWeight
        cumsumchk = [0.0]
        logger.debug("going to read all layers")
        for j, icp in enumerate(self._dcmbeam.IonControlPointSequence):
            li = LayerInfo(icp, j, cumsumchk, False, keep0)
            if 0.0 < li.mswtot or keep0:
                self._layers.append(li)
        logger.debug("survived reading all layers")
        if not is_close(mswchk, cumsumchk[0]):
            raise ValueError(
                "final cumulative msw {} != sum of spot msw {}".format(
                    mswchk, cumsumchk[0]
                )
            )
        logger.debug("survived cumulative MSW check")

    def GetAndClearWarnings(self):
        # return and clear
        w = self._warnings[:]
        self._warnings = list()
        return w

    @property
    def FinalCumulativeMetersetWeight(self):
        return float(self._dcmbeam.FinalCumulativeMetersetWeight)

    @property
    def PatientSupportAngle(self):
        return float(self._icp0.PatientSupportAngle)

    @property
    def patient_angle(self):
        return float(self._icp0.PatientSupportAngle)

    @property
    def IsoCenter(self):
        if "IsocenterPosition" in self._icp0:
            if len(self._icp0.IsocenterPosition) == 3.0:
                return [float(xyz) for xyz in self._icp0.IsocenterPosition]
            else:
                msg = "Got corrupted isocenter = '{}'; assuming [0,0,0] for now, keep fingers crossed.".format(
                    self._icp0.IsocenterPosition
                )
        else:
            msg = "No isocenter specified in treatment plan; assuming [0,0,0] for now, keep fingers crossed."
        logger.error(msg)
        self._warnings.append(msg)
        # FIXME: what to do else? Cry? Throw segfaults? Drink bad coffee?
        return [0.0, 0.0, 0.0]

    @property
    def Name(self):
        # TODO: the Name and name properties are identical, keep only one of them.
        return str(self._dcmbeam.BeamName)

    @property
    def Number(self):
        # TODO: the Number and number properties are identical, keep only one of them.
        nr = (
            str(self._index + 1)
            if self._beam_number_is_fishy
            else str(self._dcmbeam.BeamNumber)
        )
        return nr

    @property
    def name(self):
        # TODO: the Name and name properties are identical, keep only one of them.
        return str(self._dcmbeam.BeamName)

    @property
    def number(self):
        # TODO: the Number and number properties are identical, keep only one of them.
        nr = (
            str(self._index + 1)
            if self._beam_number_is_fishy
            else str(self._dcmbeam.BeamNumber)
        )
        return nr

    @property
    def RadiationType(self):
        radtype = str(self._dcmbeam.RadiationType)
        radtypeOpengate = None
        if radtype == "ION":
            ionZ = str(self._dcmbeam.RadiationAtomicNumber)
            ionA = str(self._dcmbeam.RadiationMassNumber)
            ionQ = str(self._dcmbeam.RadiationChargeState)
            radtype = "_".join(["ION", ionZ, ionA, ionQ])
            radtypeOpengate = f"ion {ionZ} {ionA}"
        return [radtype, radtypeOpengate]

    @property
    def gantry_angle(self):
        return float(self._icp0.GantryAngle)

    @property
    def TreatmentMachineName(self):
        if "TreatmentMachineName" in self._dcmbeam:
            return str(self._dcmbeam.TreatmentMachineName)
        # RayStation 8b exports anonymized treatment plans without the treatment machine name!
        if np.isclose(self.gantry_angle, 0.0):
            # FIXME: should be solved in a way that works for any clinic, not just MedAustron
            ducktape = str("IR2VBL")
        elif np.isclose(self.gantry_angle, 90.0):
            # FIXME: should be solved in a way that works for any clinic, not just MedAustron
            ducktape = str("IR2HBL")
        else:
            raise ValueError(
                "treatment machine name is missing and gantry angle {} does not enable a good guess".format(
                    self.gantry_angle
                )
            )
        msg = "treatment machine name for beam name={} number={} missing in treatment plan, guessing '{}' from gantry angle {}".format(
            self.name, self.number, ducktape, self.gantry_angle
        )
        logger.error(msg)
        self._warnings.append(msg)
        return ducktape  # ugly workaround! FIXME!

    @property
    def SnoutID(self):
        if "SnoutSequence" in self._dcmbeam:
            return str(self._dcmbeam.SnoutSequence[0].SnoutID)
        # FIXME: what to do else?
        return str("NA")

    @property
    def SnoutPosition(self):
        if "SnoutPosition" in self._dcmbeam:
            return float(self._icp0.SnoutPosition)
        # FIXME: what to do else?
        return str("NA")

    @property
    def NumberOfRangeModulators(self):
        return int(self._dcmbeam.NumberOfRangeModulators)

    @property
    def RangeModulatorIDs(self):
        if self.NumberOfRangeModulators > 0:
            return [rm.RangeModulatorID for rm in self._dcmbeam.RangeModulatorSequence]
        return list()

    @property
    def NumberOfRangeShifters(self):
        return int(self._dcmbeam.NumberOfRangeShifters)

    @property
    def RangeShifterIDs(self):
        if self.NumberOfRangeShifters > 0:
            return [str(rs.RangeShifterID) for rs in self._dcmbeam.RangeShifterSequence]
        return list()

    @property
    def NumberOfEnergies(self):
        return len(
            set(
                [icp.NominalBeamEnergy for icp in self._dcmbeam.IonControlPointSequence]
            )
        )

    @property
    def nlayers(self):
        return len(self._layers)

    @property
    def layers(self):
        return self._layers

    @property
    def nspots(self):
        return sum([l.nspots for l in self.layers])

    @property
    def mswtot(self):
        return sum([l.mswtot for l in self._layers])

    @property
    def PrimaryDosimeterUnit(self):
        return str(self._dcmbeam.PrimaryDosimeterUnit)


class BeamsetInfo(object):
    """
    This class reads a DICOM 'RT Ion Plan Storage' file and collects related information such as TPS dose files.
    It does NOT (yet) try to read a reffered structure set and/or CT images.
    This acts as a wrapper (all DICOM access on the plan file happens here). This has a few advantages over direct
    DICOM access in the other modules:
    * we can deal with different "DICOM dialects" here; some TPSs may store their plans in different ways.
    * if 'private tags' need to be taken into account then we can also do that here.
    * We can make a similar class, with the same attributes, for a treatment plan stored in a different format, e.g. for research, commissioning or QA purposes.

    Then the rest of the code can work with that in the same way.
    """

    patient_attrs = ["Patient ID", "Patient Name", "Patient Birth Date", "Patient Sex"]
    plan_req_attrs = [
        "RT Plan Label",
        "SOP Instance UID",
        "Referring Physician Name",
        "Plan Intent",
    ]
    plan_opt_attrs = ["Operators Name", "Reviewer Name", "Review Date", "Review Time"]
    plan_attrs = plan_req_attrs + plan_opt_attrs
    bs_attrs = [
        "Number Of Beams",
        "RT Plan Label",
        "Prescription Dose",
        "Target ROI Name",
        "Radiation Type",
        "Treatment Machine(s)",
    ]

    def __init__(self, rpfp):
        self._warnings = list()  # will hopefully stay empty
        self._beam_numbers_corrupt = False  # e.g. PDM does not define beam numbers
        self._rp = pydicom.dcmread(rpfp)
        self._rpfp = rpfp
        logger.debug("beamset: survived reading DICOM file {}".format(rpfp))
        self._rpdir = os.path.dirname(rpfp)
        self._rpuid = str(self._rp.SOPInstanceUID)
        self._dose_roiname = (
            None  # stays None for CT-less plans, e.g. commissioning plans
        )
        self._dose_roinumber = (
            None  # stays None for CT-less plans, e.g. commissioning plans
        )
        logger.debug("beamset: going to do some checks")
        self._chkrp()
        logger.debug("beamset: survived check, loading beams")
        self._beams = [
            BeamInfo(b, i, self._beam_numbers_corrupt)
            for i, b in enumerate(self._rp.IonBeamSequence)
        ]
        logger.debug("beamset: DONE")

    def GetAndClearWarnings(self):
        # return a copy
        for b in self._beams:
            # bwarnings = b.GetAndClearWarnings()
            for w in b.GetAndClearWarnings():
                if w not in self._warnings:
                    self._warnings.append(w)
        allw = self._warnings[:]
        self._warnings = list()
        return allw

    def __getitem__(self, k):
        if type(k) == int:
            if k >= 0 and k < len(self._beams):
                return self._beams[k]
            else:
                raise IndexError(
                    "attempt to get nonexisting beam with index {}".format(k)
                )
        for b in self._beams:
            if str(k) == b.name or str(k) == b.number:
                return b
        raise KeyError("attempt to get nonexisting beam with key {}".format(k))

    def _chkrp(self):
        if "SOPClassUID" not in self._rp:
            raise IOError("bad DICOM file {},\nmissing SOPClassUID".format(self._rpfp))
        sop_class_name = pydicom.uid.UID_dictionary[self._rp.SOPClassUID][0]
        if sop_class_name != "RT Ion Plan Storage":
            raise IOError(
                "bad plan file {},\nwrong SOPClassUID: {}='{}',\nexpecting an 'RT Ion Plan Storage' file instead.".format(
                    self._rpfp, self._rp.SOPClassUID, sop_class_name
                )
            )
        missing_attrs = list()
        for a in ["IonBeamSequence"] + self.plan_req_attrs + self.patient_attrs:
            b = a.replace(" ", "")
            if not hasattr(self._rp, b):
                missing_attrs.append(b)
        if missing_attrs:
            raise IOError(
                "bad plan file {},\nmissing keys: {}".format(
                    self._rpfp, ", ".join(missing_attrs)
                )
            )
        # self._get_rds()
        if hasattr(self._rp, "DoseReferenceSequence"):
            sequence_check(self._rp, "DoseReferenceSequence", 1, 1)
            if hasattr(self._rp.DoseReferenceSequence[0], "ReferencedROINumber"):
                self._dose_roinumber = int(
                    self._rp.DoseReferenceSequence[0].ReferencedROINumber
                )
        if self._dose_roinumber is None:
            logger.info(
                "no target ROI specified (probably because of missing DoseReferenceSequence)"
            )
        sequence_check(self._rp, "IonBeamSequence", 1, 0)
        sequence_check(self._rp, "FractionGroupSequence", 1, 1)
        frac0 = self._rp.FractionGroupSequence[0]
        sequence_check(
            frac0,
            "ReferencedBeamSequence",
            len(self._rp.IonBeamSequence),
            len(self._rp.IonBeamSequence),
        )
        number_set = set()
        for dcmbeam in self._rp.IonBeamSequence:
            nr = int(dcmbeam.BeamNumber)
            if nr < 0:
                self._beam_numbers_corrupt = True
                logger.error(
                    "CORRUPT INPUT: found a negative beam number {}.".format(nr)
                )
            if nr in number_set:
                self._beam_numbers_corrupt = True
                logger.error(
                    "CORRUPT INPUT: found same beam number {} for multiple beams.".format(
                        nr
                    )
                )
            number_set.add(nr)
        if self._beam_numbers_corrupt:
            msg = "Beam numbers are corrupt! Will override them with simple enumeration, keep fingers crossed."
            logger.error(msg)
            self._warnings.append(msg)
        logger.debug("checked planfile, looks like all attributes are available...")

    @property
    def mswtot(self):
        return sum([b.mswtot for b in self._beams])

    @property
    def name(self):
        # It looks like RTPlanLabel is for the beamset,
        # and RTPlanName is for the entire plan including possibly several beamsets.
        # the RTPlanName is not exported anymore in RayStation 8b, so let's forget about the plan name
        # some anonymization methods are butchering all useful labeling information, even the label and name of an RT plan.
        return str(self._rp.get("RTPlanLabel", "anonymized"))

    @property
    def fields(self):
        # GATE synomym for 'beams'
        return self._beams

    @property
    def beams(self):
        return self._beams

    @property
    def uid(self):
        return self._rpuid

    @property
    def beam_angles(self):
        return [str(bm.gantry_angle) for bm in self._beams]

    @property
    def beam_names(self):
        return [str(bm.Name) for bm in self._beams]

    @property
    def beam_numbers(self):
        return [str(bm.Number) for bm in self._beams]

    @property
    def Nfractions(self):
        # FIXME: some evil DICOM plan files have "NumberOfFractionsPlanned" equal to zero. Force this to be one, or is this zero somehow meaningful & useful?
        nfrac = int(self._rp.FractionGroupSequence[0].NumberOfFractionsPlanned)
        if nfrac > 0:
            return nfrac
        logger.error("Got Nfractions={} ???! Using nfrac=1 instead.".format(nfrac))
        return 1

    @property
    def target_ROI_name(self):
        return self._dose_roiname

    @target_ROI_name.setter
    def target_ROI_name(self, roiname):
        self._dose_roiname = roiname

    @property
    def target_ROI_number(self):
        return self._dose_roinumber

    @property
    def prescription_dose(self):
        if hasattr(self._rp, "DoseReferenceSequence"):
            if hasattr(self._rp.DoseReferenceSequence[0], "TargetPrescriptionDose"):
                return float(self._rp.DoseReferenceSequence[0].TargetPrescriptionDose)
        return "NA"

    @property
    def plan_label(self):
        return str(self._rp.get("RTPlanLabel", "anonymized"))

    @property
    def sanitized_plan_label(self):
        badchars = re.compile("[^a-zA-Z0-9_]")
        return re.sub(badchars, "_", self.plan_label)

    @property
    def patient_info(self):
        return dict(
            [
                (a, str(getattr(self._rp, a.replace(" ", ""))))
                for a in self.patient_attrs
            ]
        )

    @property
    def plan_info(self):
        reqs = dict(
            [
                (a, str(getattr(self._rp, a.replace(" ", ""))))
                for a in self.plan_req_attrs
            ]
        )
        opts = dict(
            [
                (
                    a,
                    (
                        "NA"
                        if not hasattr(self._rp, a.replace(" ", ""))
                        else str(getattr(self._rp, a.replace(" ", "")))
                    ),
                )
                for a in self.plan_opt_attrs
            ]
        )
        reqs.update(opts)
        return reqs

    @property
    def bs_info(self):
        info = dict(
            [
                ("Number Of Beams", str(len(self._beams))),
                ("RT Plan Label", self.plan_label),
                ("Prescription Dose", str(self.prescription_dose)),
                ("Target ROI Name", str(self.target_ROI_name)),
                (
                    "Radiation Type",
                    ", ".join(
                        set([str(beam.RadiationType[0]) for beam in self._beams])
                    ),
                ),
                (
                    "Radiation Type Opengate",
                    ", ".join(
                        set([str(beam.RadiationType[1]) for beam in self._beams])
                    ),
                ),
                (
                    "Treatment Machine(s)",
                    ", ".join(
                        set([str(beam.TreatmentMachineName) for beam in self._beams])
                    ),
                ),
            ]
        )
        dirty = self.plan_label
        sanitized = self.sanitized_plan_label
        if dirty != sanitized:
            info["Sanitized RT Plan Label"] = sanitized
        return info

    def __repr__(self):
        s = "\nPLAN\n\t" + "\n\t".join(
            ["{0:30s}: {1}".format(a, self.plan_info[a]) for a in self.plan_attrs]
        )
        s += "\nPATIENT\n\t" + "\n\t".join(
            ["{0:30s}: {1}".format(a, self.patient_info[a]) for a in self.patient_attrs]
        )
        s += "\nBEAMSET\n\t" + "\n\t".join(
            ["{0:30s}: {1}".format(a, self.bs_info[a]) for a in self.bs_attrs]
        )
        return s


class TreatmentPlanSource:
    def __init__(self, name, sim):
        self.name = name
        # self.mother = None
        self.rotation = Rotation.identity()
        self.translation = [0, 0, 0]
        self.spots = None
        self.beamline_model = None
        self.n_sim = 0
        self.sim = sim  # simulation obj to which we want to add the tp source

    def set_particles_to_simulate(self, n_sim):
        self.n_sim = n_sim

    def set_spots(self, spots):
        self.spots = spots

    def set_spots_from_rtplan(self, rt_plan_path, beam_nr=1):
        beamset = BeamsetInfo(rt_plan_path)
        gantry_angle = beamset.beam_angles[beam_nr - 1]
        spots = get_spots_from_beamset_beam(beamset, beam_nr)
        self.spots = spots
        self.rotation = Rotation.from_euler("z", gantry_angle, degrees=True)

    def set_beamline_model(self, beamline):
        self.beamline_model = beamline

    def initialize_tpsource(self, flat_generation=False, activity=False):
        # some alias
        Bq = gate.g4_units.Bq
        spots_array = self.spots
        sim = self.sim
        nSim = self.n_sim
        beamline = self.beamline_model
        self.d_nozzle_to_iso = beamline.distance_nozzle_iso
        self.d_stearMag_to_iso_x = beamline.distance_stearmag_to_isocenter_x
        self.d_stearMag_to_iso_y = beamline.distance_stearmag_to_isocenter_y

        # mapping factors between iso center plane and nozzle plane (due to steering magnets)
        cal_proportion_factor = lambda d_magnet_iso: (
            1
            if (d_magnet_iso == float("inf"))
            else (d_magnet_iso - self.d_nozzle_to_iso) / d_magnet_iso
        )
        self.proportion_factor_x = cal_proportion_factor(self.d_stearMag_to_iso_x)
        self.proportion_factor_y = cal_proportion_factor(self.d_stearMag_to_iso_y)
        tot_sim_particles = 0

        n_part_spots_V = self._sample_n_particles_spots(flat_generation=flat_generation)

        # initialize a pencil beam for each spot
        for i, spot in enumerate(spots_array):
            # if flat_generation:
            #     # simualte same number of particles for each spot
            #     nspot = nSim / len(spots_array)
            # else:
            #     # simulate a fraction of the beam particles for this spot
            #     nspot = spot.beamFraction * nSim

            nspot = n_part_spots_V[i]
            if nspot == 0:
                continue

            source = sim.add_source("IonPencilBeamSource", f"{self.name}_spot_{i}")

            # set energy
            source.energy.type = "gauss"
            source.energy.mono = beamline.get_energy(nominal_energy=spot.energy)
            source.energy.sigma_gauss = beamline.get_sigma_energy(
                nominal_energy=spot.energy
            )

            source.particle = spot.particle_name
            source.position.type = "disc"  # pos = Beam, shape = circle + sigma

            # # set mother
            # if self.mother is not None:
            #     source.attached_to = self.mother

            # POSITION:
            source.position.translation = self._get_pbs_position(spot)

            # ROTATION:
            source.position.rotation = self._get_pbs_rotation(spot)

            # add weight
            if flat_generation:
                source.weight = spot.beamFraction * len(spots_array)
                print(f"{source.weight = }")

            # set number of particles
            if activity:
                source.activity = nspot * Bq
            else:
                # nspot = np.round(nspot)
                source.n = nspot

            tot_sim_particles += nspot

            # set optics parameters
            source.direction.partPhSp_x = [
                beamline.get_sigma_x(spot.energy),
                beamline.get_theta_x(spot.energy),
                beamline.get_epsilon_x(spot.energy),
                beamline.conv_x,
            ]
            source.direction.partPhSp_y = [
                beamline.get_sigma_y(spot.energy),
                beamline.get_theta_y(spot.energy),
                beamline.get_epsilon_y(spot.energy),
                beamline.conv_y,
            ]

        self.actual_sim_particles = tot_sim_particles

    def _sample_n_particles_spots(self, flat_generation=False):
        if flat_generation:
            pdf = [1 / len(self.spots) for spot in self.spots]
        else:
            pdf = [spot.beamFraction for spot in self.spots]

        # normalize vector, to assure the probabilities sum up to 1
        pdf = pdf / np.sum(pdf)

        n_spots = len(self.spots)
        n_part_spots_V = np.zeros(n_spots)
        for i in range(int(self.n_sim)):
            bin = np.random.choice(np.arange(0, n_spots), p=pdf)
            n_part_spots_V[bin] += 1

        return n_part_spots_V

    def _get_pbs_position(self, spot):
        # (x,y) referr to isocenter plane.
        # Need to be corrected to referr to nozzle plane
        pos = [
            (spot.xiec) * self.proportion_factor_x,
            (spot.yiec) * self.proportion_factor_y,
            self.d_nozzle_to_iso,
        ]
        # Gantry angle = 0 -> source comes from +y and is positioned along negative side of y-axis
        # https://opengate.readthedocs.io/en/latest/source_and_particle_management.html

        position = (self.rotation * Rotation.from_euler("x", np.pi / 2)).apply(
            pos
        ) + self.translation

        return position

    def _get_pbs_rotation(self, spot):
        # by default the source points in direction z+.
        # Need to account for SM direction deviation and rotation thoward isocenter (270 deg around x)
        # then rotate of gantry angle
        rotation = [0.0, 0.0, 0.0]
        beta = np.arctan(spot.yiec / self.d_stearMag_to_iso_y)
        alpha = np.arctan(spot.xiec / self.d_stearMag_to_iso_x)
        rotation[0] = -np.pi / 2 + beta
        rotation[2] = -alpha

        # apply gantry angle
        spot_rotation = (
            self.rotation * Rotation.from_euler("xyz", rotation)
        ).as_matrix()

        return spot_rotation



import copy
from pathlib import Path
from typing import Optional, List
from difflib import get_close_matches
from functools import wraps

from box import Box
import sys

from .exception import (
    fatal,
    warning,
    GateDeprecationError,
    GateFeatureUnavailableError,
    GateImplementationError,
)
from .definitions import (
    __gate_list_objects__,
    __gate_dictionary_objects__,
    __one_indent__,
)
from .decorators import requires_fatal
import traceback
from .logger import logger


def print_call_location():
    stack = traceback.extract_stack()
    # Get the caller of this function
    filename, lineno, _, _ = stack[-3]
    msg = f"Called from file: {filename}, line: {lineno}"
    return msg


# Singletons
class MetaSingletonFatal(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in MetaSingletonFatal._instances:
            MetaSingletonFatal._instances[cls] = super(
                MetaSingletonFatal, cls
            ).__call__(*args, **kwargs)
            return MetaSingletonFatal._instances[cls]
        else:
            fatal(
                f"You are trying to create another instance of {cls.__name__}, but an instance already exists "
                f"in this process. Only one instance per process can be created. "
                f"Please open a new python session and run the simulation again. "
            )


class GateSingletonFatal(metaclass=MetaSingletonFatal):
    pass


# base class for objects handling user input
class MetaUserInfo(type):
    _created_classes = {}

    def __call__(cls, *args, **kwargs):
        process_cls(cls)
        return super(MetaUserInfo, type(cls)._created_classes[cls]).__call__(
            *args, **kwargs
        )


def process_cls(cls):
    """This factory function is meant to process classes inheriting from GateObject.
    GateObject.__process_this__() handles the user infos of GateObjects
    and general functionality common to GateObjects.
    Subsets of classes, e.g. the actor classes,
    can specialize the factoring steps in their own __process_this__() class method.
    """
    if not cls.has_been_processed():
        cls.__process_this__()


def wrap_init_method(cls):
    """This is a factory function to process classes which inherit from GateObject.
    It is called from the main factory function process_cls().
    This function wraps and reattaches the __init__ method of this class, if it implements one.
    The wrapped __init__ first calls the "original" __init__ and subsequently the method
    __finalize_init__, which has a base implementation in GateObject,
    in case the __init__ is the furthest down in the inheritance chain.
    The method __finalize_init__ is needed to allow GateObject.__setattr__ to check for invalid attribute setting.
    """
    # Get the __init__ method as the class implements it
    original_init = cls.__dict__.get("__init__")
    # if it is implemented, i.e. present in __dict__, wrap it
    if original_init is not None:
        # define a closure
        @wraps(original_init)
        def wrapped_init(self, *args, **kwargs):
            # original_init is the __init__ captured in the closure
            original_init(self, *args, **kwargs)
            # figure out in which class the __init__ method is implemented.
            # It could be in some super class with respect to the instance self.
            class_to_which_original_init_belongs = vars(
                sys.modules[original_init.__module__]
            )[original_init.__qualname__.split(".")[0]]
            # Now figure out which is the "last" class in the inheritance chain
            # (with respect to the instance self)
            # which implements an __init__ method. Plus the children which do not implement an __init__
            classes_up_to_first_init_in_mro = []
            for c in type(self).mro():
                classes_up_to_first_init_in_mro.append(c)
                if "__init__" in c.__dict__:
                    # found an __init__, so __init__ methods in classes further up the inheritance tree
                    # should not call the __finalize_init__ method
                    break
            # Now check if the class in which the __init__ we are wrapping is implemented
            # is among the previously extracted classes.
            # If that is the case, the call to this __init__ will be the last one to terminate
            # in the chain if super().__init__() calls.
            # In other words, we are at the very end of __init__, including calls to super classes,
            # and it is time to call __finalize_init__
            if class_to_which_original_init_belongs in classes_up_to_first_init_in_mro:
                self.__finalize_init__()

        # reattach the wrapped __init__ to the class, so it is used instead of the original one.
        setattr(cls, "__init__", wrapped_init)


# Utility function for object creation
def check_property_name(name):
    if len(name.split()) > 1:
        fatal("Invalid property name: f{name}. Should not contain spaces.")


def check_property(property_name, value, defaultvalue):
    msg = (
        lambda required: f"{property_name} requires a {required}, found {type(value).__name__}."
    )
    if isinstance(defaultvalue, str) and not isinstance(value, str):
        raise Exception(msg("string"))
    elif type(defaultvalue) is bool and type(value) is not bool:
        raise Exception(msg("bool"))
    elif type(defaultvalue) is bool and type(value) is bool:
        return
    elif isinstance(defaultvalue, (int, float, complex)) and (
        not isinstance(value, (int, float, complex)) or isinstance(value, bool)
    ):
        raise Exception(msg("number"))


def digest_user_info_defaults(cls):
    inherited_user_info_defaults = {}
    # loop through MRO backwards so that inherited classes
    # are processed first
    for c in cls.mro():
        # check if the class actual define user_info_defaults
        # note: hasattr() would not work because it would yield the attribute from the
        # base class if the inherited class does not define user_info_defaults
        if "user_info_defaults" in c.__dict__:
            # Make sure there are no duplicate user info items.
            # First check for user info defaults that override the one from super classes and exclude them
            inherited_user_info_defaults_keys_override_true = [
                k
                for k, v in inherited_user_info_defaults.items()
                if "override" in v[1] and v[1]["override"] is True
            ]
            user_info_defaults_to_be_added = dict(
                [
                    (k, v)
                    for k, v in c.user_info_defaults.items()
                    if k not in inherited_user_info_defaults_keys_override_true
                ]
            )
            if set(user_info_defaults_to_be_added.keys()).isdisjoint(
                set(inherited_user_info_defaults.keys())
            ):
                inherited_user_info_defaults.update(user_info_defaults_to_be_added)
            else:
                fatal(
                    f"Implementation error. "
                    f"Duplicate user info defined for class {cls}."
                    f"Found {list(user_info_defaults_to_be_added.keys())}."
                    f"Base classes already contain {list(inherited_user_info_defaults.keys())}. "
                )
        else:
            # Ensure that the class defines an empty dictionary
            # so that the user_info_defaults from the base class will not show up.
            try:
                setattr(c, "user_info_defaults", {})
            except TypeError:
                # TypeError is thrown if the class is 'object'
                pass
    cls = add_properties_to_class(cls, inherited_user_info_defaults)
    cls.inherited_user_info_defaults = inherited_user_info_defaults

    return cls


def add_properties_to_class(cls, user_info_defaults):
    """Add user_info defaults as properties to class if not yet present."""
    for p_name, default_value_and_options in user_info_defaults.items():
        _ok = False
        if (
            isinstance(default_value_and_options, tuple)
            and len(default_value_and_options) == 2
        ):
            default_value = default_value_and_options[0]
            options = default_value_and_options[1]
        else:
            s = (
                f"*** DEVELOPER WARNING ***"
                f"User info defaults possibly not implemented correctly for class {cls}.\n"
                "The value for each user info item in the user info dictionary \n"
                "should be a tuple where the first item is the default value, \n"
                "and the second item is a (possibly empty) dictionary of options.\n"
            )
            fatal(s)
            options = None  # remove warning from IDE
            default_value = None  # remove warning from IDE
        if "deprecated" not in options:
            if not hasattr(cls, p_name):
                check_property_name(p_name)
                setattr(
                    cls, p_name, _make_property(p_name, default_value, options=options)
                )

                try:
                    expose_items = options["expose_items"]
                except KeyError:
                    expose_items = False
                if expose_items is True:
                    # expose_items can only be used on dictionary-type user infos
                    # try to get keys and fail of impossible (=not dict type)
                    try:
                        for item_name, item_default_value in default_value.items():
                            check_property_name(item_name)
                            if not hasattr(cls, item_name):
                                setattr(
                                    cls,
                                    item_name,
                                    _make_property(
                                        item_name,
                                        item_default_value,
                                        container_dict=p_name,
                                    ),
                                )
                            else:
                                fatal(
                                    f"Duplicate user info {item_name} defined for class {cls}. Check also base classes or set 'expose_items=False."
                                )
                    except AttributeError:
                        fatal(
                            "Option 'expose_items=True' not available for default_user_info {p_name}."
                        )
    return cls


def _make_property(property_name, default_value, options=None, container_dict=None):
    """Return a property that stores the user_info item in a
    dictionary which is an attribute of the object (self).

    """

    if options is None:
        options = {}

    # @property
    def prop_getter(self):
        if container_dict is None:
            if "getter_hook" in options:
                return options["getter_hook"](self, self.user_info[property_name])
            else:
                return self.user_info[property_name]
        else:
            return self.user_info[container_dict][property_name]

    read_only = options.get("read_only", False)
    if read_only is False:

        # @prop.setter
        def prop_setter(self, value):
            if "deactivated" in options and options["deactivated"] is True:
                if value != self.inherited_user_info_defaults[property_name][0]:
                    raise GateFeatureUnavailableError(
                        f"The user input parameter {property_name} "
                        f"is currently deactivated and cannot be set."
                    )
            if "deprecated" in options:
                raise GateDeprecationError(options["deprecated"])
            if container_dict is None:
                if "setter_hook" in options:
                    value_to_be_set = options["setter_hook"](self, value)
                else:
                    value_to_be_set = value
                if "allowed_values" in options:
                    if value_to_be_set not in options["allowed_values"]:
                        fatal(
                            f"Object {self.name} received illegal value {value_to_be_set} "
                            f"for user input {property_name}. Allow values are: {options['allowed_values']}."
                        )
                self.user_info[property_name] = value_to_be_set
            else:
                self.user_info[container_dict][property_name] = value

    else:
        prop_setter = None

    prop_doc = make_docstring_for_user_info(property_name, default_value, options)
    prop = property(fget=prop_getter, fset=prop_setter, doc=prop_doc)

    return prop


def convert_default_value_to_string(default_value):
    default_value_str = str(default_value)
    default_value_str = default_value_str.replace("\n", ", ")
    return default_value_str


def make_docstring_for_user_info(name, default_value, options):
    begin_of_line = "  * "
    docstring = f"{name}"
    if "deprecated" in options:
        docstring += ":\n\n"
        docstring += f"{begin_of_line}Deprecated: {options['deprecated']}\n"
    else:
        if "required" in options and options["required"] is True:
            docstring += " (must be provided)"
        if "read_only" in options and options["read_only"] is True:
            docstring += " (set internally, i.e. read-only)"
        docstring += ":\n\n"
        docstring += f"{begin_of_line}Default value: {convert_default_value_to_string(default_value)}\n"
        if "allowed_values" in options:
            docstring += f"{begin_of_line}Allowed values: {options['allowed_values']}\n"
        if "doc" in options:
            docstring += f"{begin_of_line}Description: {options['doc']}\n"
    return docstring


def help_on_user_info(obj):
    if hasattr(obj, "__user_info_doc__"):
        print(obj.__user_info_doc__)
    elif type(obj) is property and hasattr(obj, "__doc__"):
        print(obj.__doc__)
    else:
        raise GateImplementationError(
            f"No __user_info_doc__ or __doc__ available for {obj}. "
        )


def restore_instance_after_pickling(cls, attributes):
    # In the context of sub-processing and pickling,
    # the following line makes sure the class is processed by the function
    # which sets handles the user_info definitions
    # before the class is used to create a new object instance.
    # Otherwise, the new instance would lack the user_info properties.
    process_cls(cls)
    # this is just conventional unpickling logic:
    obj = cls.__new__(cls)
    obj.__dict__.update(attributes)
    return obj


# class GateObject(metaclass=MetaUserInfo):
class GateObject:
    """This is the base class used for all objects that handle user input in GATE.

    The class is assumed to be processed by process_cls(), either explicitly
    or via the metaclass MetaUserInfo, before any instances of the class are created.
    Some class attributes, e.g. inherited_user_info_defaults, are created as part of this processing.
    """

    # hints for IDE
    name: str
    inherited_user_info_defaults: dict

    user_info_defaults = {"name": (None, {"required": True})}

    @classmethod
    def __get_user_info_docstring__(cls):
        docstring = f"User input parameters and default values:\n\n"
        for k, v in sorted(cls.inherited_user_info_defaults.items()):
            default_value = v[0]
            options = v[1]
            docstring += "* "
            docstring += make_docstring_for_user_info(k, default_value, options)
            docstring += "\n"
        docstring += "\n"
        return docstring

    @classmethod
    def __get_docstring__(cls):
        if cls.__doc__ is not None:
            docstring = cls.__doc__
            docstring += "\n\n"
        else:
            docstring = ""
        cls.__user_info_doc__ = cls.__get_user_info_docstring__()
        docstring += cls.__user_info_doc__
        # docstring += 20 * "*"
        docstring += "\n"
        return docstring

    @classmethod
    def has_been_processed(cls):
        """The class attribute inherited_user_info_defaults is exclusively set by the  factory mechanism.
        Therefore, if this class does not yet have an attribute inherited_user_info_defaults,
        it means that it has not been processed yet.
        Note: we cannot use hasattr(cls, 'inherited_user_info_defaults')
        because it would potentially find the attribute from already processed base classes.
        Therefore, we must use cls.__dict__ which contains only attributes of the specific cls object.
        """

        return "inherited_user_info_defaults" in cls.__dict__

    @classmethod
    def __process_this__(cls):
        cls.__process_user_info_defaults__()
        cls.__doc__ = cls.__get_docstring__()

    @classmethod
    def __process_user_info_defaults__(cls):
        """Internal interface class method used e.g. by __reduce__().
        Certain subgroups of classes, e.g. actor classes, may implement this differently
        without the need to adapt the __reduce__()  method.

        This factory method is meant to process classes inheriting from GateObject.
        It digests the user info parametrisation from all classes in the inheritance tree
        and enhances the __init__ method, so it calls the __finalize_init__ method at the
        very end of the __init__ call, which is required to check for invalid attribute setting.
        """
        if cls.has_been_processed():
            return

        try:
            digest_user_info_defaults(cls)
        except AttributeError:
            raise GateImplementationError(
                "Looks like you are calling process_cls on a class "
                "that does not inherit from GateObject."
            )
        # the class attribute known_attributes is needed by the __setattr__ method of GateObject
        cls.known_attributes = set()
        # enhance the __init__ method to ensure __finalize_init__ is called at the end
        wrap_init_method(cls)

    def __new__(cls, *args, **kwargs):
        # We need to make sure the class has been processed.
        # This is relevant for classes defined by the user,
        # e.g. a CustomTranslationChanger (test030)
        # where process_cls() is not explicitly called as it is the case
        # for classes implemented within the GATE package
        process_cls(cls)
        new_instance = super(GateObject, cls).__new__(cls)
        return new_instance

    def __init__(self, *args, simulation=None, **kwargs) -> None:
        self._simulation = simulation
        # keep the internal number of raised warnings (for debug)
        self.number_of_warnings = 0
        self._temporary_warning_cache = []
        # prefill user info with defaults
        self.user_info = Box(
            [
                (k, copy.deepcopy(v[0]))
                for k, v in self.inherited_user_info_defaults.items()
            ]
        )
        # now iterate over them and check if kwargs provide user-specific values
        for k, v in self.inherited_user_info_defaults.items():
            options = v[1]
            if k in kwargs:
                user_info_value = kwargs[k]
                if "setter_hook" in options:
                    user_info_value = options["setter_hook"](self, user_info_value)
                self.user_info[k] = user_info_value
                kwargs.pop(k)
            else:
                if "required" in options.keys() and options["required"] is True:
                    fatal(
                        f"No value provided for argument '{k}', but required when constructing a {type(self).__name__} object."
                    )
        mro = type(self).__mro__
        parent = mro[mro.index(__class__) + 1]
        if type(parent).__name__ != "pybind11_type":
            try:
                super().__init__(*args, **kwargs)
            except TypeError:
                raise TypeError(
                    f"There was a problem "
                    f"while trying to create the {type(self).__name__} called {self.name}. \n"
                    f"Check if you have provided unknown keyword arguments. "
                    f"You provided: {list(kwargs.keys())}. \n"
                    f"Hint: The user input parameters of {type(self).__name__} are: "
                    f"{list(self.inherited_user_info_defaults.keys())}.\n"
                )

    @property
    def simulation(self):
        return self._simulation

    @simulation.setter
    def simulation(self, sim):
        sim.warnings.extend(self._temporary_warning_cache)
        self._temporary_warning_cache = []
        self._simulation = sim

    def __str__(self):
        ret_string = (
            f"GateObject {type(self).__name__} named {self.name} "
            f"with the following parameters:\n"
        )
        for k, v in self.user_info.items():
            if k != "name":
                ret_string += f"{__one_indent__}{k}"
                if "deprecated" in self.inherited_user_info_defaults[k][1]:
                    ret_string += " (deprecated)"
                ret_string += f":\n{2 * __one_indent__}{v}\n"
        ret_string += "***\n"
        return ret_string

    def __getstate__(self):
        """Method needed for pickling. May be overridden in inheriting classes."""
        for k, v in self.__dict__.items():
            if "engine" in k and v is not None:
                warning(
                    f"Potential bug: Object {self.name} of type {self.type_name} "
                    f"had a reference to {k} that was not None when being pickled. "
                    f"That should not be the case! \n"
                    f"Info for developers: \n"
                    f"Probably, a line self.{k}=None is needed in {self.type_name}.close()."
                )
        if self.simulation is not None and self.simulation.verbose_getstate:
            warning(
                f"__getstate__() called in object '{self.name}' of type {self.type_name}."
            )

        # Note: returning a copy, e.g. via dict([(k, v) for k, v in self.__dict__.items()])
        # instead of returning self.__dict__ leads to infinite recursion during pickling.
        # Reason:
        # Many objects hold circular references, e.g. Simulation has a reference to VolumeManager and vice versa
        # The pickle module can handle this as long as the objects are identical.
        # If we return a copy of dict here, pickle is unable to understand and handle the circularity.
        return self.__dict__

    def __setstate__(self, d):
        """Method needed for pickling. May be overridden in inheriting classes."""
        self.__dict__ = d

    def __reduce__(self):
        """This method is called when the object is pickled.
        Usually, pickle works well without this custom __reduce__ method,
        but objects handling user_infos need a custom __reduce__ to make sure
        the properties linked to the user_infos are properly created.

        The return arguments are:
        1) A callable used to create the instance when unpickling
        2) A tuple of arguments to be passed to the callable in 1
        3) The dictionary of the object's properties to be passed to the __setstate__ method (if defined)
        """
        state_dict = self.__getstate__()
        return (
            restore_instance_after_pickling,
            (
                self.__class__,
                state_dict,
            ),  # the second argument will be reassigned to __dict__
            # when re-instantiating the class, but we must pass state_dict and not
            # self.__dict__ because self.__dict__ might contain non-pickable objects,
            # while __getstate__() should by design return a pickable representation
            state_dict,
        )

    def __setattr__(self, key, value):
        # raise an error if the user tries to set an attribute
        # associated with a deprecated user input parameter
        if (
            key in self.inherited_user_info_defaults
            and "deprecated" in self.inherited_user_info_defaults[key][1]
        ):
            raise GateDeprecationError(
                self.inherited_user_info_defaults[key][1]["deprecated"]
            )

        # check if the attribute is known, otherwise warn the user
        known_attributes = type(self).__dict__.get("known_attributes")
        if known_attributes is None:
            raise GateImplementationError(
                f"Did not find 'known_attributes' in the {self.type_name}. "
                f"Has the class correctly been processed by process_cls()?"
            )
        if len(known_attributes) > 0:
            if key not in known_attributes:
                msg = f'For object "{self.name}", attribute "{key}" is not known. Maybe a typo?\n'
                msg += print_call_location() + "\n"
                close_matches = get_close_matches(key, known_attributes)
                if len(close_matches) > 0:
                    msg_close_matches = (
                        "Did you mean: " + " or ".join(close_matches) + "\n"
                    )
                    msg += msg_close_matches
                known_attr = ", ".join(
                    str(a) for a in known_attributes if not a.startswith("_")
                )
                msg += f"Known attributes of this object are: {known_attr}"
                self.warn_user(msg)
                self.number_of_warnings += 1
        super().__setattr__(key, value)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()
        return False

    def __finalize_init__(self):
        """This method is called once all attributes have been defined, i.e.
        at the end of the __init__ method. It defines the set of known_attribues that will
        be used to detect errors when the user tries to use a new attribute
        or misspells an attribute, e.g. box.mohter instead of box.mother.
        """

        # we define this at the class-level
        type(self).known_attributes = set(dir(self))

    def __add_to_simulation__(self):
        """Hook method which can be called by managers.
        Specific classes can use this to implement actions to be taken
        when an object is being added to the simulation,
        e.g. adding a certain actor implies switching on certain physics options.
        """
        pass

    @property
    def type_name(self):
        return str(type(self).__name__)

    def close(self):
        """Dummy implementation for inherited classes which do not implement this method."""
        if "simulation" in self.__dict__ and self.simulation is not None:
            if self.simulation.verbose_close:
                warning(
                    f"close() called in object '{self.name}' of type {type(self).__name__}."
                )

    def release_g4_references(self):
        """Dummy implementation for inherited classes which do not implement this method."""
        pass

    def configure_like(self, other_obj):
        for k in self.user_info.keys():
            if k not in ["name", "_name"]:
                try:
                    self.user_info[k] = other_obj.user_info[k]
                except KeyError:
                    pass

    def to_dictionary(self):
        d = {
            "user_info": {},
            "object_type": str(type(self).__name__),
            "object_type_full": str(type(self)),
            "class_module": type(self).__module__,
            "i_am_a_gate_object": True,
        }
        for k, v in self.user_info.items():
            d["user_info"][k] = recursive_userinfo_to_dict(v)
        return d

    def from_dictionary(self, d):
        try:
            if d["object_type_full"] != str(type(self)):
                fatal(
                    f"Error while populating object named {self.name}: "
                    f"Incompatible dictionary associated with object type {d['object_type']}"
                )
        except KeyError:
            fatal(
                f"Error while populating object named {self.name}: "
                "The provided dictionary does not contain any info about the object type."
            )
        for k in self.user_info.keys():
            if k in d["user_info"]:
                if hasattr(self, k):
                    # get the class property associate with the user info end check if it has a setter
                    # otherwise, it is read-only
                    if getattr(type(self), k).fset is not None:
                        setattr(self, k, d["user_info"][k])
                else:
                    if "deprecated" not in self.inherited_user_info_defaults[k][1]:
                        warning(
                            f"Could not find user info {k} while populating object {self.name} "
                            f"of type {type(self).__name__} from dictionary. "
                            f"The reason could be that the user parameter is marked as deprecated. "
                            f"In that case, simply ignore the warning. "
                        )

    def warn_user(self, message):
        # If this GateObject does not (yet) have a reference to the simulation,
        # we store the warning in a temporary cache
        # (will be registered later to the simulation's warning cache)
        if self.simulation is None:
            self._temporary_warning_cache.append(message)
        # if possible, register the warning directly
        else:
            self.simulation._user_warnings.append(message)
        warning(message)


class DynamicGateObject(GateObject):
    # hints for IDE
    dynamic_params: Optional[List]

    user_info_defaults = {
        "dynamic_params": (
            None,
            {
                "doc": "Dictionary of dictionaries, where each dictionary specifies how the parameters "
                "of this object should evolve over time during the simulation. "
                "You cannot set this parameter directly. "
                "Instead, use the 'add_dynamic_parametrisation()' method of your object."
                "If None, the object is static (default).",
                "read_only": True,
            },
        )
    }

    @property
    def is_dynamic(self):
        if self.dynamic_params is None:
            return False
        elif len(self.dynamic_params) == 0:
            return False
        else:
            return True

    @property
    def dynamic_user_info(self):
        return [
            k
            for k in self.user_info
            if "dynamic" in self.inherited_user_info_defaults[k][1]
            and self.inherited_user_info_defaults[k][1]["dynamic"] is True
        ]

    @requires_fatal("simulation")
    def process_dynamic_parametrisation(self, params):
        # create a dictionary to store params which to not correspond to dynamic user info
        # i.e. extra parameters for auxiliary purpose
        extra_params = {}
        extra_params["auto_changer"] = params.pop(
            "auto_changer", True
        )  # True of key not found (default)
        if extra_params["auto_changer"] not in (False, True):
            fatal(
                f"Received wrong value type for 'auto_changer': got {type(extra_params['auto_changer'])}, "
                f"expected: True or False."
            )
        for k in set(params).difference(set(self.dynamic_user_info)):
            extra_params[k] = params.pop(k)
        # apply params which are functions to the timing intervals of the simulation to get the sample quantities
        for k, v in params.items():
            if callable(v):
                params[k] = v(self.simulation.run_timing_intervals)

        return params, extra_params

    def check_if_dynamic_params_match_run_timing_intervals(self):
        # check that the length of all parameter lists match the simulation's timing intervals
        params_with_incorrect_length = []
        for params in self.dynamic_params.values():
            for k, v in params.items():
                if k in self.dynamic_user_info and len(v) != len(
                    self.simulation.run_timing_intervals
                ):
                    params_with_incorrect_length.append((k, len(v)))
            if len(params_with_incorrect_length) > 0:
                s = (
                    "The length of the following dynamic parameters "
                    "does not match the number of timing intervals of the simulation:\n"
                )
                for p in params_with_incorrect_length:
                    s += f"{p[0]}: {p[1]}\n"
                s += (
                    f"The simulation's timing intervals are: {self.simulation.run_timing_intervals} and "
                    f"can be adjusted via the simulation parameter 'run_timing_intervals'. "
                )
                fatal(s)

    def _add_dynamic_parametrisation_to_userinfo(self, params, name):
        """This base class implementation only acts as a setter.
        Classes inheriting from this class should implement an
        add_dynamic_parametrisation() method which actually does something
        with the parameters and then call super().add_dynamic_parametrisation().
        Inheriting classes should avoid calling this method directly.
        """
        if name not in self.user_info["dynamic_params"]:
            self.user_info["dynamic_params"][name] = params
        else:
            fatal(
                f"A dynamic parametrisation with name {name} already exists in volume '{self.name}'."
            )

    def add_dynamic_parametrisation(self, name=None, **params):
        if self.user_info["dynamic_params"] is None:
            self.user_info["dynamic_params"] = {}
        processed_params, extra_params = self.process_dynamic_parametrisation(params)
        processed_params["extra_params"] = extra_params
        # if the user provided no name, create one
        if name is None:
            name = f"parametrisation_{len(self.dynamic_params)}"
        self._add_dynamic_parametrisation_to_userinfo(processed_params, name)
        # issue a debugging message
        s = f"Added the following dynamic parametrisation to {type(self).__name__} '{self.name}': \n"
        for k, v in processed_params.items():
            s += f"{k}: {v}\n"
        logger.debug(s)

    def create_changers(self):
        # this base class implementation is here to keep inheritance intact.
        return []


class GateUserInputSwitchDict(Box):
    """
    NOT USED YET!

    Specialized version of a Box (dict) to represent a dictionary with boolean switches.

    The switches handled by the object need to be defined when the object is created
    via a dictionary passed as argument 'default_switches'.
    No switches can be added later, nor can switches be removed.
    Switch values are automatically converted to Bool if possible.
    """

    def __init__(self, default_switches, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._switches = tuple(default_switches.keys())
        for k, v in default_switches.items():
            self[k] = v

    def __setitem__(self, key, value):
        # Do not allow setting entries under the key '_switches' which is used as special attribute
        if key == "_switches":
            raise KeyError(f"Keyword '_switches' is not allowed.")
        # only try setting the item if the key is known
        elif key in self._switches:
            try:
                value_bool = bool(value)
            except ValueError:
                raise ValueError(
                    "You must provide a boolean (or compatible) input, i.e. True or False."
                )
            super().__setitem__(key, value_bool)
        else:
            raise KeyError(
                "You cannot add additional switches. You can only turn on/off existing switches."
            )

    def __delitem__(self, key):
        """The 'del' operator applied on items is blocked so no entries can be removed."""
        raise NotImplementedError("You cannot remove switches.")

    def __setattr__(self, key, value):
        """Make sure to by-pass the __setattr__ method from the Box class for the key '_switches'
        because Box would otherwise turn this into an entry in self, but we want it to be a pure attribute.
        """
        if key == "_switches":
            object.__setattr__(self, key, value)
        else:
            super().__setattr__(key, value)


def recursive_userinfo_to_dict(obj):
    """Walk recursively across entries of user_info and convert to appropriate structure.
    Dictionary-like structures are mapped to dictionary and walked across recursively.
    List-like structures are mapped to lists and walked across recursively.
    GateObject-like objects are converted through their to_dictionary() method.
    All other input (presumably common data types including numpy structures) is left untouched.
    """

    if isinstance(obj, __gate_dictionary_objects__):
        ret = {}
        for k, v in obj.items():
            ret[k] = recursive_userinfo_to_dict(v)
    elif isinstance(obj, __gate_list_objects__):
        ret = []
        for e in obj:
            ret.append(recursive_userinfo_to_dict(e))
    elif isinstance(obj, GateObject):
        ret = obj.to_dictionary()
    else:
        ret = obj
    return ret


def find_paths_in_gate_object_dictionary(go_dict, only_input_files=False):
    paths = []
    for ui_name, ui in go_dict["user_info"].items():
        new_paths = find_all_paths(ui)
        if only_input_files is True:
            options = _get_user_info_options(
                ui_name, go_dict["object_type"], go_dict["class_module"]
            )
            try:
                consider_this = options["is_input_file"]
            except KeyError:
                consider_this = False
        else:
            consider_this = True
        if consider_this is True:
            paths.extend(new_paths)
    return paths


def recursively_search_object(obj, condition=(lambda x: True)):
    found_objects = []
    if condition(obj) is True:
        found_objects.append(obj)
    if isinstance(obj, __gate_dictionary_objects__):
        for v in obj.values():
            found_objects.extend(recursively_search_object(v, condition=condition))
    if isinstance(obj, __gate_list_objects__):
        for e in obj:
            found_objects.extend(recursively_search_object(e, condition=condition))
    return found_objects


def find_all_gate_objects(dct):
    def condition(obj):
        try:
            ret = obj["i_am_a_gate_object"]
        except (KeyError, TypeError, IndexError):
            ret = False
        return ret

    return recursively_search_object(dct, condition=condition)


def find_all_paths(dct):
    def condition(obj):
        return isinstance(obj, (Path))

    return recursively_search_object(dct, condition=condition)


def _get_user_info_options(user_info_name, object_type, class_module):
    """Utility function to retrieve the options associated with a user info given the class name,
    the module in which the class is defined, and the name of the user info.
    """
    # this is a workaround because dynamically created actor output classes
    # cannot be found in the modules dict and will raise an Attribute error
    # a better way would be to implement the to_dictionary method in a way
    # that it also stores a list of input files
    # so that we do not need to do this recursive search here
    try:
        cls = getattr(sys.modules[class_module], object_type)
    except AttributeError:
        return {}
    try:
        options = cls.inherited_user_info_defaults[user_info_name][1]
    except KeyError:
        fatal(f"Could not find user info {user_info_name} in {object_type}. ")
        options = None  # remove warning from IDE
    return options


def create_gate_object_from_dict(dct):
    """Function to (re-)create an object derived from GateObject based on a dictionary.

    Used as part of the deserialization chain, when reading simulations stored as JSON file.
    """
    if "class_module" not in dct:
        fatal(
            "Error while trying to create GateObject from dictionary: Incompatible dictionary"
        )
    obj = getattr(sys.modules[dct["class_module"]], dct["object_type"])(
        name=dct["user_info"]["name"]
    )
    return obj


process_cls(GateObject)
process_cls(DynamicGateObject)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as nm670
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test028_ge_nm670_spect", output_folder="test028"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "qt"
    sim.number_of_threads = 1
    sim.check_volumes_overlap = False
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    kBq = 1000 * Bq

    # world size
    sim.world.size = [1 * m, 1 * m, 1 * m]
    sim.world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect, colli, crystal = nm670.add_spect_head(sim, "spect", debug=False)
    nm670.rotate_gantry(spect, 20 * cm, 0)

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [15 * cm, 15 * cm, 15 * cm]
    waterbox.material = "G4_WATER"
    blue = [0, 1, 1, 1]
    waterbox.color = blue

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False

    sim.physics_manager.set_production_cut(
        volume_name="world", particle_name="all", value=10 * mm
    )  # all means proton, electron, positron, gamma

    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="gamma",
        value=0.1 * mm,
    )
    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="electron",
        value=0.1 * mm,
    )
    sim.physics_manager.set_production_cut(
        volume_name="spect",
        particle_name="positron",
        value=0.1 * mm,
    )

    # default source for tests
    activity = 100 * kBq
    beam1 = sim.add_source("GenericSource", "beam1")
    beam1.attached_to = waterbox.name
    beam1.particle = "gamma"
    beam1.energy.mono = 140.5 * keV
    beam1.position.type = "sphere"
    beam1.position.radius = 3 * cm
    beam1.position.translation = [0, 0, 0 * cm]
    beam1.direction.type = "momentum"
    beam1.direction.momentum = [0, 1, 0]
    beam1.activity = activity / sim.number_of_threads

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # start simulation
    sim.run()

    # stat
    gate.exception.warning("Compare stats")
    print(stats)
    print(f"Number of runs was {stats.counts.runs}. Set to 1 before comparison")
    stats.counts.runs = 1  # force to 1
    stats_ref = utility.read_stat_file(paths.gate_output / "stat1.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.02)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test010_generic_source", output_folder="test010"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output

    # useful units
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    deg = gate.g4_units.deg
    mm = gate.g4_units.mm
    m = gate.g4_units.m
    cm = gate.g4_units.cm

    # set the world size like in the Gate macro
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # add a simple volume
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
    waterbox.material = "G4_WATER"

    # test sources
    source = sim.add_source("GenericSource", "source1")
    source.particle = "gamma"
    source.activity = 10000 * Bq / sim.number_of_threads
    source.position.type = "sphere"
    source.position.radius = 5 * mm
    source.position.translation = [-3 * cm, 30 * cm, -3 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, -1, 0]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV

    source = sim.add_source("GenericSource", "source2")
    source.particle = "proton"
    source.activity = 10000 * Bq / sim.number_of_threads
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.position.translation = [6 * cm, 5 * cm, -30 * cm]
    # source.position.rotation = Rotation.from_euler('x', 45, degrees=True).as_matrix()
    source.position.rotation = Rotation.identity().as_matrix()
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.energy.type = "gauss"
    source.energy.mono = 140 * MeV
    source.energy.sigma_gauss = 10 * MeV

    source = sim.add_source("GenericSource", "s3")
    source.particle = "proton"
    source.activity = 10000 * Bq / sim.number_of_threads
    source.position.type = "box"
    source.position.size = [4 * cm, 4 * cm, 4 * cm]
    source.position.translation = [8 * cm, 8 * cm, 30 * cm]
    source.direction.type = "focused"
    source.direction.focus_point = [1 * cm, 2 * cm, 3 * cm]
    source.energy.type = "gauss"
    source.energy.mono = 140 * MeV
    source.energy.sigma_gauss = 10 * MeV

    source = sim.add_source("GenericSource", "s4")
    source.particle = "proton"
    source.activity = 10000 * Bq / sim.number_of_threads
    source.position.type = "box"
    source.position.size = [4 * cm, 4 * cm, 4 * cm]
    source.position.translation = [-3 * cm, -3 * cm, -3 * cm]
    # source.position.rotation = Rotation.from_euler('x', 45, degrees=True).as_matrix()
    source.position.rotation = Rotation.identity().as_matrix()
    source.direction.type = "iso"
    source.energy.type = "gauss"
    source.energy.mono = 80 * MeV
    source.energy.sigma_gauss = 1 * MeV

    # actors
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")

    # src_info = sim.add_actor('SourceInfoActor', 'src_info')
    # src_info.filename = 'output/sources.root'

    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test010.mhd"
    dose.attached_to = waterbox
    dose.size = [50, 50, 50]
    dose.spacing = [4 * mm, 4 * mm, 4 * mm]

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")
    # sim.g4_commands_after_init.append("/run/verbose 2")
    # sim.g4_commands_after_init.append("/event/verbose 2")
    # sim.g4_commands_after_init.append("/tracking/verbose 1")

    # start simulation
    sim.run()

    # print
    print("Simulation seed:", sim.current_random_seed)

    # get results
    print(stats_actor)
    print(dose)

    # gate_test10
    # Gate mac/main.mac
    # Current version is two times slower :(
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    print("-" * 80)
    is_ok = utility.assert_stats(stats_actor, stats_ref, tolerance=0.05)
    is_ok = is_ok and utility.assert_images(
        paths.gate_output / "output-Edep.mhd",
        dose.edep.get_output_path(),
        stats_actor,
        tolerance=30,
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )

    utility.test_ok(is_ok)



# Version Information
# Python:   3.9.18
# Pandas:   2.2.2
# NumPy:    1.24.1

import opengate as gate
import opengate.tests.utility as tu
from opengate.contrib.optical.optigan import OptiGAN
import platform

import os

if __name__ == "__main__":
    paths = tu.get_default_test_paths(__file__, output_folder="test075_optigan")

    # create simulation
    sim = gate.Simulation()
    sim.g4_verbose = True
    sim.output_dir = paths.output
    # sim.random_seed = 600

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    eV = gate.g4_units.eV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq

    # add a material database
    print(f"Inside the test file - {paths.data}")
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # set the world size like in the Gate macro
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]

    # add a simple crystal volume
    crystal = sim.add_volume("Box", "crystal")
    crystal.size = [3 * mm, 3 * mm, 20 * mm]
    crystal.translation = [0 * cm, 0 * cm, 0 * cm]
    crystal.material = "BGO"
    crystal.set_production_cut("electron", 0.1 * mm)

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.energy_range_min = 10 * eV
    sim.physics_manager.energy_range_max = 1 * MeV
    sim.physics_manager.special_physics_constructors.G4OpticalPhysics = True

    source = sim.add_source("GenericSource", "gamma1")
    source.particle = "gamma"
    source.energy.mono = 0.511 * MeV
    source.activity = 10 * Bq
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.position.translation = [0 * cm, 0 * cm, 2.2 * cm]

    # filter : remove opticalphoton
    fe = sim.add_filter("ParticleFilter", "fe")
    fe.particle = "opticalphoton"
    fe.policy = "reject"

    phsp_actor = sim.add_actor("PhaseSpaceActor", "Phase")
    phsp_actor.attached_to = crystal
    # hc.output = paths.output / "test075_simulation_optigan_with_random_seed.root"
    phsp_actor.attributes = [
        "Position",
        "PostPosition",
        "PrePosition",
        "ParticleName",
        "TrackCreatorProcess",
        "EventKineticEnergy",
        "KineticEnergy",
        "PDGCode",
        "ParentID",
        "EventID",
        "TrackID",
    ]

    phsp_actor.steps_to_store = "first"
    phsp_actor.output_filename = "test075_simulation_optigan_with_random_seed_600.root"

    # add a kill actor to the crystal
    ka = sim.add_actor("KillActor", "kill_actor2")
    ka.attached_to = crystal
    ka.filters.append(fe)

    sim.user_hook_after_run = gate.userhooks.user_hook_dump_material_properties
    sim.run()

    # *** 2 ways to use OptiGAN ***

    # Option 1:
    # Call OptiGAN class with input_phsp_actor set to the correct phase space actor
    # Output will be saved in the folder specified via sim.output_dir
    device = "auto"
    if platform.system() == "Darwin":
        device = "cpu"
    optigan = OptiGAN(input_phsp_actor=phsp_actor, torch_device=device)

    # -------

    # Option 2:
    # alternatively, without a simulation,
    # you can pass the keyword argument root_file_path
    # optigan = OptiGAN(root_file_path=hc.get_output_path())

    # In that case, if you want the optigan output to be saved under sim.output_dir
    # you need to set the simulation
    # optigan.simulation = sim
    # otherwise, output will be saved under the current directory where the script resides

    # use
    # help(optigan)
    # for an explanation of the input parameters
    # -------

    # Run run_optigan method of Optigan class to get outputs
    # Option create_output_graphs: Generates distribution graphs for each output feature.
    #   -> Use only with low source activity, i.e. few events in the phase levels may cause memory issues.
    optigan.run_optigan(create_output_graphs=False)

    is_ok = all(t is True for t in sim.user_hook_log)
    tu.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test029_volume_time_rotation_helpers as test029
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test029"
    )

    # create the main simulation object
    sim = gate.Simulation()

    # create sim without AA
    test029.create_simulation(sim, False, paths, "_trot1")

    # for later reference, get the actors that were created by the helper function above
    proj_actor = sim.actor_manager.get_actor("Projection")
    stats = sim.actor_manager.get_actor("Stats")

    # initialize & start
    sim.run()

    """
    # use to create the (fake) reference for test029_volume_time_rotation_2.py
    import itk
    scaling = 0.90
    img = itk.imread(paths.output_ref / "proj029.mhd")
    arr = itk.array_view_from_image(img)
    arr *= scaling
    itk.imwrite(img, paths.output_ref / "proj029_scaled.mhd")
    """

    # -------------------------
    gate.exception.warning("Compare stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "stats029.txt")
    print(
        f"Number of steps was {stats.counts.steps}, forced to the same value (because of angle acceptance). "
    )
    stats.counts.steps = stats_ref.counts.steps  # force to id
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.02)
    print(is_ok)

    gate.exception.warning("Compare images")
    # read image and force change the offset to be similar to old Gate
    is_ok = (
        utility.assert_images(
            paths.output_ref / "proj029.mhd",
            proj_actor.get_output_path("counts"),
            stats,
            tolerance=59,
            ignore_value_data2=0,
            axis="x",
            sum_tolerance=2,
            apply_ignore_mask_to_sum_check=False,  # reproduce legacy behavior
        )
        and is_ok
    )
    print(is_ok)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.userhooks import check_production_cuts

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test023")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12332567
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # iec phantom
    iec_phantom = gate_iec.add_iec_phantom(sim)

    # default source for tests
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 50 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 1 * cm
    source.direction.type = "iso"
    source.activity = 10000 * Bq

    # filter : keep gamma
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    fp = sim.add_filter("ParticleFilter", "fp")
    fp.particle = "e-"
    fk = sim.add_filter("KineticEnergyFilter", "fk")
    fk.energy_min = 100 * keV

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test023_iec_phantom.mhd"
    dose.attached_to = "iec"
    dose.size = [100, 100, 100]
    dose.spacing = [2 * mm, 2 * mm, 2 * mm]
    dose.filters = [fp, fk]

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True
    stat.filters.append(f)

    print(stat)
    print(dose)
    print("Filters: ", sim.filter_manager)
    print(sim.filter_manager.dump())

    # change physics
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMZ"
    sim.physics_manager.global_production_cuts.all = 0.1 * mm
    sim.user_hook_after_init = check_production_cuts

    # start simulation
    sim.run(start_new_process=True)

    # tests
    f = paths.output_ref / "test023_stats_iec_phantom.txt"
    stats_ref = utility.read_stat_file(f)
    is_ok = utility.assert_stats(stat, stats_ref, 0.12)
    is_ok = is_ok and utility.assert_images(
        paths.output_ref / "test023_iec_phantom.mhd",
        dose.edep.get_output_path(),
        stat,
        tolerance=102,
        sum_tolerance=28,
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate.sources.phidsources as phid
import click
import matplotlib.pyplot as plt

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("rad_name", nargs=1)
@click.option("--output", "-o", default=None, help="output file")
def go(rad_name, output):
    nuclide = phid.print_phid_info(rad_name)
    print(nuclide)

    fig, ax = nuclide.plot(label_pos=0.66)
    if output:
        print(f"Output plot in {output}")
        fig.savefig(output)
    else:
        plt.show()


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()
